# -*- coding: UTF-8 -*-
from ga2.engine.engine import *
from ga2.engine.uielement import UIElement


class UnityEngine(GameEngine):
    def __init__(self, address, port):
        super(UnityEngine, self).__init__(address, port)
        self._callback_socket = None
        self._callback_functions = {}
        self._callback_thread = None

    def engine_type(self):
        return "Unity"

    def _parse_path(self, path):
        if path is None:
            return None

        nodeInfos = []
        nodes = path.split("/")

        if len(nodes) == 0:
            return None
        elif nodes[0] == '':
            pathNode = {"name": "/", "index": -1, "txt": "", "img": "", "regex": ""}
            nodeInfos.append(pathNode)
            nodes = nodes[1:]
        name_re = re.compile(r"(?P<name>[^[{]+)")
        txt_re = re.compile(r"(txt\s*=\s*(?P<txt>[^,\}]*))")
        img_re = re.compile(r"(img\s*=\s*(?P<img>[^,\}]*))")
        index_re = re.compile(r"\[(?P<index>\d+)\]")
        regex_re = re.compile(r"\{\{(?P<regex>.+)\}\}")
        for node in nodes:
            result = regex_re.search(node)
            if result:
                regex = result.groupdict().get("regex", "")
                name = ""
            else:
                regex = ""
                result = name_re.search(node)
                if result:
                    name = result.groupdict().get("name", "")
                else:
                    name = ""

            result = txt_re.search(node)
            if result:
                txt = result.groupdict().get("txt", "")
            else:
                txt = ""

            result = img_re.search(node)
            if result:
                img = result.groupdict().get("img", "")
            else:
                img = ""
            result = index_re.search(node)
            if result:
                index = result.groupdict().get("index", "-1")
                index = int(index)
            else:
                index = -1
            pathNode = {"name": name, "index": index, "txt": txt, "img": img, "regex": regex}
            nodeInfos.append(pathNode)
        return nodeInfos

    def find_elements_path(self, path):
        """
        表达式匹配获取符合的所有节点
        第一个节点"/"代表从根节点，如果不是代表可以从任意位置开始


        index：代表节点的位置。
             parent
           /      \
         index[0]  index[1]

         name:代表节点名称，"*"代表任意名称
         text:代表节点的文字内容（当前GameObject的挂载的Compenent）
         img:图片名称,sprite或者texture名称


         /Canvas/Panel/Button
         Canvas[0]{txt="hello"}/Button
         /Canvas/Panel/*[1]/Button
        :param path:
            需要查找的路径,/Canvas/Button[0]{txt=Button,img=img}
        :Usage:
            >>>button=engine.find_elements_path('/Canvas/Panel/Button')
            >>>button2=engine.find_elements_path('/Canvas/Panel/*[1]/Button')
            >>>button3=engine.find_elements_path('Canvas/Button{txt=hello}')
        :return:包含instance的Elements对象列表
            example:
            [{"objectName":"/Canvas/Panel/Button",
            "Instance":4257741},{"objectName":"/Canvas/Panel/Button",
            "Instance":4257742}]
        """
        nodeInfos = self._parse_path(path)
        if nodeInfos is None:
            raise WeTestInvaildArg("Error path")

        ret = self.socket.send_command(Commands.FIND_ELEMENT_PATH, nodeInfos)
        elements = []
        for e in ret:
            element = Element(e["name"], e["instance"])
            elements.append(element)

        return elements

    def find_elements_by_component(self, name):
        """
            通过GameObject.FindObjectsOfType(Type.GetType(name))查找对应的GameObject.

            Type.GetType获取不到的话，默认会尝试name,Assembly-CSharp, Version=0.0.0.0, Culture=neutral, PublicKeyToken=null
            及name, Assembly-CSharp-firstpass, Version=0.0.0.0, Culture=neutral, PublicKeyToken=null
        :param name:
            查找的类型的，AssemblyQualifiedName，如WetestManager,U3DAutomation, Version=0.0.0.0, Culture=neutral, PublicKeyToken=null
        :return:
            当前界面中所有挂载改界面的
        """
        ret = self.socket.send_command(Commands.FIND_ELEMENTS_COMPONENT, [name])
        elements = []
        for e in ret:
            element = Element(e["name"], e["instance"])
            elements.append(element)
        return elements

    def get_element_text(self, element):
        """
            获取GameObject文字信息
            NGUI控件则获取UILable、UIInput、GUIText组件上的文字信息
            UGUI控件则获取Text、GUIText组件上的问题信息
        :param element: 查找到的GameObject

        :Usage:
            >>>element=engine.find_element('/Canvas/Panel/Button')
            >>>text=engine.get_element_text(element)
        :return:文字内容
        :raises WeTestInvaildArg,WeTestRuntimeError
        """
        if element is None:
            raise WeTestInvaildArg("Invalid Instance")
        ret = self.socket.send_command(Commands.GET_ELEMENT_TEXT, element.instance)
        return ret

    def get_element_image(self, element):
        """
            获取GameObject图片名称
            NGUI控件则获取UITexture、UISprite、SpriteRenderer组件上的图片名称
            UGUI控件则获取Image、RawImage、SpriteRenderer组件上的图片名称
        :param element: 查找到的GameObject

        :Usage:
            >>>element=engine.find_element('/Canvas/Panel/Button')
            >>>image=engine.get_element_image(element)
        :return:图片名称
        :rtype: str
        :raises WeTestInvaildArg,WeTestRuntimeError
        """
        if element is None:
            raise WeTestInvaildArg("Invalid Instance")
        ret = self.socket.send_command(Commands.GET_ELEMENT_IMAGE, element.instance)
        return ret

    def get_registered_handlers(self):
        """
            获取用户当前注册的自定义的函数集合
        :Usage:
            >>>function_names=engine.get_registered_handlers()
        :return: []注册的自定义函数名称序列
        :raise WeTestInvaildArg,WeTestRuntimeError
        """
        ret = self.socket.send_command(Commands.GET_REGISTERED_HANDLERS)
        return ret

    def call_registered_handler(self, name, args):
        """
            调用指定的注册的函数，并返回返回值
        :param name:已经注册的函数名称
        :param args:传入函数中的参数，一个不超过1024个字符的参数
        :param timeout:超时时间(超时后直接返回，SDK中对应的函数可能还是会执行完，不会中断)
        :return:
            自定义注册函数的返回值
        :raise WeTestInvaildArg,WeTestRuntimeError
        """
        ret = self.socket.send_command(Commands.CALL_REGISTER_HANDLER, {"name": name, "args": args})
        return ret

    def get_element_bound(self, element):
        """
        获取GameObject在屏幕上的位置和长宽高
        :param element: 查找到的GameObject

        :Usage:
            >>>
            >>>element=engine.find_element('/Canvas/Panel/Button')
            >>>bound=engine.get_element_bound(element)
        :return:屏幕中的位置（x,y），左上角为原点，及长宽
            examples:
            {"x":0.5，
            "y":0.5.0,
            "width":0.2
            "height":0.1}
        :rtype: ElementBound
        :raises WeTestInvaildArg,WeTestRuntimeError
        """
        if element is None:
            raise WeTestInvaildArg("Invaild Instance,element is None")

        ret = self._get_elements_bound([element])
        print(ret)
        if ret:
            result = ret[0]
            if not result["existed"]:
                return None
            else:
                return ElementBound(result["x"], result["y"], result["width"], result["height"], result["visible"])
        return None

    def _get_elements_bound(self, elements):
        send_params = [e.instance for e in elements]
        ret = self.socket.send_command(Commands.GET_ELEMENTS_BOUND, send_params)
        return ret

    def get_touchable_uielements(self, params=None):

        """
            获取当前界面的可点击节点的列表
        :return:
            [(element1,point1),(element2,point2)]
            [({"object_name":"/Canvas/Panel/Button","Instance":4257741},{"x":250,"y":300}),.....]
        :raise WeTestRuntimeError
        """
        ret = self.socket.send_command(Commands.GET_UI_INTERACT_STATUS, params)
        elements = []
        for i in range(0, len(ret["elements"])):
            node = ret["elements"][i]
            element = Element(node["name"], node["instanceid"])
            bound = ElementBound(node["bound"]["x"], node["bound"]["y"], node["bound"]["fWidth"],
                                 node["bound"]["fHeight"])
            elements.append(UIElement(element, bound))
        return elements


    def get_touchable_elements_bound(self, params=None):

        """
            获取当前界面的可点击节点的列表
        :Usage:
            >>>buttons=engine.get_touchable_elements_bound()
        :return:
            [(element1,point1),(element2,point2)]
            [({"object_name":"/Canvas/Panel/Button","Instance":4257741},{"x":250,"y":300}),.....]
        :raise WeTestRuntimeError
        """
        ret = self.socket.send_command(Commands.GET_UI_INTERACT_STATUS, params)
        elements = []
        for i in range(0, len(ret["elements"])):
            node = ret["elements"][i]
            element = Element(node["name"], node["instanceid"])
            bound = ElementBound(node["bound"]["x"], node["bound"]["y"], node["bound"]["fWidth"],
                                 node["bound"]["fHeight"])
            elements.append((element, bound))
        return elements

    # android only
    def _inject_touch_actions(self, actions, timeout=20):

        """
            发送touch序列号，touch事件结束之后，才会返回。同步函数
        :param actions:
        :return:
        """
        ret = self.socket.send_command(Commands.HANDLE_TOUCH_EVENTS, actions, timeout)
        return ret

    #android only
    def swipe_android(self, start_element, end_element, steps=20, duration=1000):
        """

        :param start_element:
        :param end_element:
        :param steps:步数，通过控制步数能够控制滑动时长和滑动平滑度，每一步走的耗时在step_sleep为5ms左右。100步时长>500毫。无法做到精确控制，调试确认
        :param step_sleep:每个步骤的间隔时长，单位毫秒

        :Usage:
            >>>start_element=engine.find_element('/Canvas/Panel/Button')
            >>>end_element=engine.find_element('/Canvas/Panel/Button2')
            >>>engine.swipe(start_element,end_element,5000)
        :return: WeTestRuntimeError,WeTestInvaildArg
        """
        start_x = 0
        start_y = 0
        end_x = 0
        end_y = 0
        if start_element and isinstance(start_element, Element) and end_element and isinstance(end_element, Element):
            ret = self._get_elements_bound([start_element, end_element])
            if ret and len(ret) == 2:
                result = ret[0]
                if not result["existed"] or not result["visible"]:
                    reason = "StartElement = {0} not existed or unvisible".format(start_element)
                    raise WeTestInvaildArg(reason)
                else:
                    start_x = result["x"] + result["width"] / 2
                    start_y = result["y"] + result["height"] / 2

                result = ret[1]
                if not result["existed"] or not result["visible"]:
                    reason = "EndElement = {0} not existed or unvisible".format(start_element)
                    raise WeTestInvaildArg(reason)
                else:
                    end_x = result["x"] + result["width"] / 2
                    end_y = result["y"] + result["height"] / 2
        elif start_element and isinstance(start_element, ElementBound) and end_element and isinstance(end_element,
                                                                                                      ElementBound):
            start_x = start_element.x + start_element.width / 2
            start_y = start_element.y + start_element.height / 2
            end_x = end_element.x + end_element.width / 2
            end_y = end_element.y + end_element.height / 2
        else:
            reason = "Input start_element = {0},end_element = {1},vaild argument is Element or ElementBound".format(
                start_element, end_element)
            raise WeTestInvaildArg(reason)

        return self.swipe_position(start_x, start_y, end_x, end_y, steps, duration=duration)

    # android only
    def swipe_and_press_android(self, start_x, start_y, end_x, end_y, steps, duration, step_sleep=5):
        """
        滑动并在结束的地方一直按压
        :param start_x: 起始位置x，绝对坐标
        :param start_y: 起始位置y，绝对坐标
        :param end_x: 结束位置x,绝对坐标
        :param end_y:结束位置y，绝对坐标
        :param steps: 滑动中间的步骤数，每一步的间隔为5ms，可以用于控制滑动速度和平滑度
        :param step_sleep:每个步骤的间隔时长，单位毫秒
        :param duration: 结束位置按压时间，单位是毫秒ms

        :Usage:
            >>>engine.swipe_and_press(start_x,start_y,end_x,end_y,50,10000,step_sleep=200)
        :raise WeTestInvaildArg,WeTestRuntimeError
        """
        if steps <= 0:
            reason = "steps = {0} is invaild, steps must more than 0".format(steps)
            raise WeTestInvaildArg(reason)
        start_x = int(start_x)
        start_y = int(start_y)
        end_x = int(end_x)
        end_y = int(end_y)
        actions = [{"x": start_x, "y": start_y, "sleeptime": 0, "type": TouchEvent.ACTION_DOWN}]

        x_distance = (end_x - start_x) * 1.0 / steps
        y_distance = (end_y - start_y) * 1.0 / steps
        move_x, move_y = start_x, start_y

        for i in range(steps - 1):
            move_x += x_distance
            move_y += y_distance
            actions.append(
                {"x": int(move_x), "y": int(move_y), "sleeptime": step_sleep, "type": TouchEvent.ACTION_MOVE})

        move_x += x_distance
        move_y += y_distance
        actions.append(
            {"x": int(move_x), "y": int(move_y), "sleeptime": duration, "type": TouchEvent.ACTION_MOVE})

        actions.append({"x": int(end_x), "y": int(end_y), "sleeptime": 0, "type": TouchEvent.ACTION_UP})

        self._inject_touch_actions(actions, timeout=60)
        return True

    #android only
    def input_android(self, locator, text):
        """

        :param locator: is an element locator
        :param text: input的内容

        :Usage:
            >>>element=engine.find_element("/Canvas/Panel/input")
            >>>old_txt=engine.input(element,"new text")
        :return: "hello world"input控件上原有的内容
        :raise WeTestInvaildArg,WeTestRuntimeError
        """
        if locator and isinstance(locator, Element):
            result = self.socket.send_command(Commands.SET_INPUT_TEXT, {"instance": locator.instance, "content": text})
            return result
        else:
            reason = "Input locator = {0},text = {1},vaild argument is Element or ElementBound".format(locator, text)
            raise WeTestInvaildArg(reason)

    def get_element_world_bound(self, elements):
        """
            查找节点对应的世界坐标。世界坐标包含，节点的中心位置的x,y,z坐标，及物体离中心点在在x,y,z轴上的大小。
            具体详见：http://docs.unity3d.com/ScriptReference/bound.html
        :param elements: 一个Element或者Element[]，节点应该至少包含Renderer、MeshFilter或Colider组件
        :return: WorldBound[]

        :Usage:
            >>>hero1=engine.find_element("hero1")
            >>>hero2=engine.find_element("hero2")
            >>>bound=engine.get_element_world_bound([hero1,hero2])
        :raise WeTestInvaildArg，WeTestRuntimeError
        """
        if elements is None:
            raise WeTestInvaildArg("Invaild Instance")
        if isinstance(elements, Element):
            elements = [elements]

        if len(elements) == 0:
            raise WeTestInvaildArg("Invaild Instance,search node is error")

        req = [e.instance for e in elements]
        ret = self.socket.send_command(Commands.GET_ELEMENT_WORLD_BOUND, req)

        world_bound = []
        for res in ret:
            world_bound = WorldBound(res["id"], res["existed"])
            if world_bound.existed:
                world_bound.center_x = res["centerX"]
                world_bound.center_y = res["centerY"]
                world_bound.center_z = res["centerZ"]
                world_bound.extents_x = res["extentsX"]
                world_bound.extents_y = res["extentsY"]
                world_bound.extents_z = res["extentsZ"]
                world_bound.append(world_bound)

        return world_bound

    def get_component_field(self, element, component, attribute):
        """
            通过反射的方式获取到GameObject上面组件的属性值。反射查看是否存在Property或者Field名称为attribute
        :param element:已经找到的GameObject对象
        :param component:组件名称
        :param attribute:字段或者属性名称
        :return: 字段或者属性名称toString()之后的string值

        :Usage:
            >>>elements = engine.find_elements_path("Sample/Text")
            >>>res = engine.get_component_field(elements[0], "Text", "text")
        :raise WeTestInvaildArg，WeTestRuntimeError
        """
        if element is None or component is None or attribute is None:
            raise WeTestInvaildArg("Invaild Instance")

        ret = self.socket.send_command(Commands.GET_OBJECT_FIELD,
                                       {"instance": element.instance, "comopentName": component,
                                        "attributeName": attribute})

        return ret

    def set_camera(self, camera):
        ret = self.socket.send_command(Commands.SET_CAMERA_NAME, [camera])
        return ret

    def get_component_methods(self, element, component):
        """
        通过反射获取组件的方法
        :param element:已经找到的GameObject对象
        :param component:组件名称
        :return:方法的描述，包括方法名称，返回值类型，参数类型
        :Usage:
            >>>element = engine.find_element("Sample")
            >>>methods = engine.get_component_methods(element, "ReflectionTest")
        :raise WeTestInvaildArg，WeTestRuntimeError
        """
        if element is None or component is None:
            raise WeTestInvaildArg("Invaild Instance")

        ret = self.socket.send_command(Commands.GET_COMPONENT_METHODS,
                                       {"instance": element.instance, "comopentName": component})
        return ret

    def call_component_method(self, element, component, method, params):
        """
        通过反射调用组件上的方法
        :param element:已经找到的GameObject对象
        :param component:组件名称
        :param method:方法名称
        :param params:方法参数数组
        :return:方法的返回
        :Usage:
            >>>element = engine.find_element("Sample")
            >>>params = [5, "Hello World"]
            >>>result = engine.call_component_method(element, "ReflectionTest", "TestReflection", params)
        :raise WeTestInvaildArg，WeTestRuntimeError
        """
        if element is None or component is None or method is None:
            raise WeTestInvaildArg("Invaild Instance")

        ret = self.socket.send_command(Commands.CALL_COMPONENT_MOTHOD,
                                       {"instance": element.instance, "comopentName": component,
                                        "methodName": method, "parameters": params})
        return ret


    def click_android(self, locator):
        """
            点击一个button, checkbox or radio button.
            'locator' is an Element locator or ElementBound
            如果是一个Element会先去查找Element的位置，然后点击中心点
            如果是一个ElementBound则会直接点击，x,y的位置
        :param locator:
            点击的节点为，节点位置
        :Usage:
            >>>element=engine.find_element('/Canvas/Panel/Button')
            >>>bound=engine.get_element_bound(element)
            >>>engine.click_android(bound)

            >>>element=engine.find_element('/Canvas/Panel/Button2')
            >>>engine.click_android(element)
        :raise WeTestInvaildArg,WeTestRuntimeError
        """
        if locator is None:
            return
        if isinstance(locator, Element):
            try:
                bound = self.get_element_bound(locator)
                if bound:
                    return self.click_position_android(bound.x + bound.width / 2, bound.y + bound.height / 2)
            except WeTestRuntimeError as e:
                logger.error("Get element({0}) bound faild {1}".format(locator, e.message))
            return False
        elif isinstance(locator, ElementBound) and locator:
            try:
                return self.click_position_android(locator.x + locator.width / 2, locator.y + locator.height / 2)
            except WeTestRuntimeError:
                logger.error("Get element({0}) bound faild".format(locator))
            return False
        else:
            reason = "Click locator = {0},vaild argument is Element or ElementBound".format(locator)
            raise WeTestInvaildArg(reason)

    def press_android(self, locator, press_time):
        """
            长按动作
        :param locator:is an element locator
        :param press_time:按压时间，单位为毫秒

        :Usage:
            >>>element=engine.find_element('/Canvas/Panel/Button')
            >>>bound=engine.get_element_bound(element)
            >>>engine.press_android(bound,1000)

            >>>element=engine.find_element('/Canvas/Panel/Button2')
            >>>engine.press_android(element,5000)
        :raises WeTestInvaildArg,WeTestRuntimeError
        """

        if locator is None:
            return
        if isinstance(locator, Element):
            try:
                bound = self.get_element_bound(locator)
                return self.press_position(bound.x + bound.width / 2, bound.y + bound.height / 2, press_time)
            except WeTestRuntimeError:
                logger.error("Get element({0}) bound faild".format(locator))
            return False
        elif isinstance(locator, ElementBound):
            try:
                return self.press_position(locator.x + locator.width / 2, locator.y + locator.height / 2, press_time)
            except WeTestRuntimeError:
                logger.error("Get element({0}) bound faild".format(locator))
            return False
        else:
            reason = "Press locator = {0},vaild argument is Element or ElementBound".format(locator)
            raise WeTestInvaildArg(reason)


    def click_position_android(self, x, y):
        """
            在屏幕的某个位置进行点击
        :param x: 150 在屏幕x=150的位置进行点击
        :param y: 200 在屏幕y=200的位置进行点击

        :Usage:
            >>>engine.click_position_android(100,300)

        :raise WeTestRuntimeError
        """
        x = int(x)
        y = int(y)
        actions = [{"x": x, "y": y, "sleeptime": 50, "type": TouchEvent.ACTION_DOWN},
                   {"x": x, "y": y, "sleeptime": 0, "type": TouchEvent.ACTION_UP}]
        logger.debug("click {0}".format(actions))
        self._inject_touch_actions(actions)
        return True

    # only supported for android and no-il2cpp games
    # def game_script_init(self, path):
    #     """
    #         将gametestlib.dll push到手机上的/data/local/tmp目录下
    #         然后调用gametestlib.dll中的GameTest.Test.init()方法
    #     :param path: gametestlib.dll， inject c# script
    #     :return:
    #     """
    #     logger.debug("push c# test script : {0}".format(path))
    #     result = AdbTool().cmd_wait("push", path, "/data/local/tmp/gametestlib.dll")
    #     logger.debug("push result : {0}".format(result))
    #
    #     ret = self.socket.send_command(Commands.LOAD_TEST_LIB)
    #     return ret
    #
    # def register_game_callback(self, name, func):
    #     if not name or not func:
    #         raise WeTestInvaildArg("Name or Function can't be None")
    #     if not self._callback_socket or not self._callback_functions:
    #         self._callback_socket = SocketClient(self.address, self.port)
    #         event = threading.Event()
    #         self._callback_thread = RPCReceiveThread(self._callback_socket, self._callback_functions, event)
    #         self._callback_thread.start()
    #
    #     self._callback_functions[name] = func
    #     self._callback_socket.send_package(Commands.PRC_SET_METHOD, name)
    #     return self._callback_thread.get_result()



# -*- coding: UTF-8 -*-

import cv2
from ga2.common.utils import callLog
from ga2.device.deviceOrientation import *
from ga2.device.screenShoter import ScreenShoter
import numpy

class DefaultScreenshoter(ScreenShoter):

    def __init__(self, wdaconnector ):
        self.__wda_connector = wdaconnector

    @callLog
    def screenshot(self, localpath=None, portrait=True, height=None):

        pngdata = self.__wda_connector.wda_client.screenshot(localpath)
        pngarr = numpy.fromstring(pngdata, numpy.uint8)
        origin_image = cv2.imdecode(pngarr, cv2.IMREAD_UNCHANGED)
        scaled_image = self.resize_screenshot(origin_image, height) if height else origin_image
        scaled_width, scaled_height = scaled_image.shape[1], scaled_image.shape[0]
        rotated_image = scaled_image
        if portrait:
            orientation = self.__wda_connector.get_session().orientation
            orientation = WDA_ORIENTATION_MAP[orientation]
            if orientation == (
                    DeviceOrientation.LANDSCAPERIGHT or orientation == DeviceOrientation.LANSCAPE) and scaled_height > scaled_width:  # in case the origin picture is  portrait
                orientation = DeviceOrientation.PORTRAIT
            rotated_image = self.rotate_to_portrait(scaled_image, orientation)
        if localpath:
            cv2.imwrite(filename=localpath, img=rotated_image)
        return rotated_image

    def resize_screenshot(self, origin, length):
        if length is None:
            return origin
        display_size = self.__wda_connector.get_session().window_size()
        if display_size[0] > display_size[1]:
            width = int(length * display_size[1] / display_size[0])
            return cv2.resize(origin, (length, width))
        else:
            width = int(length * display_size[0] / display_size[1])
            return cv2.resize(origin, (width, length))





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python wrapper for Android uiautomator tool."""

import sys
import os
import subprocess
import time
import itertools
import json
import hashlib
import socket
import re
import collections
import xml.dom.minidom
import logging
from ga2.common import imapUtf7
import six

logger=logging.getLogger(__name__)

DEVICE_PORT = int(os.environ.get('UIAUTOMATOR_DEVICE_PORT', '9008'))
LOCAL_PORT = int(os.environ.get('UIAUTOMATOR_LOCAL_PORT', '9008'))

if 'localhost' not in os.environ.get('no_proxy', ''):
    os.environ['no_proxy'] = "localhost,%s" % os.environ.get('no_proxy', '')

try:
    import urllib2
except ImportError:
    import urllib.request as urllib2
try:
    from httplib import HTTPException
except:
    from http.client import HTTPException
try:
    if os.name == 'nt':
        import urllib3
except:  # to fix python setup error on Windows.
    pass

__author__ = "Xiaocong He"
__all__ = ["device", "Device", "rect", "point", "Selector", "JsonRPCError"]


def U(x):
    if sys.version_info.major == 2:
        return x.decode('utf-8') if type(x) is str else x
    elif sys.version_info.major == 3:
        return x


def param_to_property(*props, **kwprops):
    if props and kwprops:
        raise SyntaxError("Can not set both props and kwprops at the same time.")

    class Wrapper(object):

        def __init__(self, func):
            self.func = func
            self.kwargs, self.args = {}, []

        def __getattr__(self, attr):
            if kwprops:
                for prop_name, prop_values in kwprops.items():
                    if attr in prop_values and prop_name not in self.kwargs:
                        self.kwargs[prop_name] = attr
                        return self
            elif attr in props:
                self.args.append(attr)
                return self
            raise AttributeError("%s parameter is duplicated or not allowed!" % attr)

        def __call__(self, *args, **kwargs):
            if kwprops:
                kwargs.update(self.kwargs)
                self.kwargs = {}
                return self.func(*args, **kwargs)
            else:
                new_args, self.args = self.args + list(args), []
                return self.func(*new_args, **kwargs)

    return Wrapper


class JsonRPCError(Exception):
    def __init__(self, code, message):
        self.code = int(code)
        self.message = message

    def __str__(self):
        return "JsonRPC Error code: %d, Message: %s" % (self.code, self.message)


class JsonRPCMethod(object):
    if os.name == 'nt':
        try:
            pool = urllib3.PoolManager()
        except:
            pass

    def __init__(self, url, method, timeout=30):
        self.url, self.method, self.timeout = url, method, timeout

    def __call__(self, *args, **kwargs):
        if args and kwargs:
            raise SyntaxError("Could not accept both *args and **kwargs as JSONRPC parameters.")
        data = {"jsonrpc": "2.0", "method": self.method, "id": self.id()}
        if args:
            data["params"] = args
        elif kwargs:
            data["params"] = kwargs
        jsonresult = {"result": ""}
        if os.name == "nt":
            res = self.pool.urlopen("POST",
                                    self.url,
                                    headers={"Content-Type": "application/json"},
                                    body=json.dumps(data).encode("utf-8"),
                                    timeout=self.timeout)
            jsonresult = json.loads(res.data.decode("utf-8"))
        else:
            result = None
            try:
                req = urllib2.Request(self.url,
                                      json.dumps(data).encode("utf-8"),
                                      {"Content-type": "application/json"})
                result = urllib2.urlopen(req, timeout=self.timeout)
                jsonresult = json.loads(result.read().decode("utf-8"))
            finally:
                if result is not None:
                    result.close()
        if "error" in jsonresult and jsonresult["error"]:
            raise JsonRPCError(
                jsonresult["error"]["code"],
                "%s: %s" % (jsonresult["error"]["data"]["exceptionTypeName"], jsonresult["error"]["message"])
            )
        return jsonresult["result"]

    def id(self):
        m = hashlib.md5()
        m.update(("%s at %f" % (self.method, time.time())).encode("utf-8"))
        return m.hexdigest()


class JsonRPCClient(object):
    def __init__(self, url, timeout=30, method_class=JsonRPCMethod):
        self.url = url
        self.timeout = timeout
        self.method_class = method_class

    def __getattr__(self, method):
        return self.method_class(self.url, method, timeout=self.timeout)


class Selector(dict):
    """The class is to build parameters for UiSelector passed to Android device.
    """
    __fields = {
        "text": (0x01, None),  # MASK_TEXT,
        "textContains": (0x02, None),  # MASK_TEXTCONTAINS,
        "textMatches": (0x04, None),  # MASK_TEXTMATCHES,
        "textStartsWith": (0x08, None),  # MASK_TEXTSTARTSWITH,
        "className": (0x10, None),  # MASK_CLASSNAME
        "classNameMatches": (0x20, None),  # MASK_CLASSNAMEMATCHES
        "description": (0x40, None),  # MASK_DESCRIPTION
        "descriptionContains": (0x80, None),  # MASK_DESCRIPTIONCONTAINS
        "descriptionMatches": (0x0100, None),  # MASK_DESCRIPTIONMATCHES
        "descriptionStartsWith": (0x0200, None),  # MASK_DESCRIPTIONSTARTSWITH
        "checkable": (0x0400, False),  # MASK_CHECKABLE
        "checked": (0x0800, False),  # MASK_CHECKED
        "clickable": (0x1000, False),  # MASK_CLICKABLE
        "longClickable": (0x2000, False),  # MASK_LONGCLICKABLE,
        "scrollable": (0x4000, False),  # MASK_SCROLLABLE,
        "enabled": (0x8000, False),  # MASK_ENABLED,
        "focusable": (0x010000, False),  # MASK_FOCUSABLE,
        "focused": (0x020000, False),  # MASK_FOCUSED,
        "selected": (0x040000, False),  # MASK_SELECTED,
        "packageName": (0x080000, None),  # MASK_PACKAGENAME,
        "packageNameMatches": (0x100000, None),  # MASK_PACKAGENAMEMATCHES,
        "resourceId": (0x200000, None),  # MASK_RESOURCEID,
        "resourceIdMatches": (0x400000, None),  # MASK_RESOURCEIDMATCHES,
        "index": (0x800000, 0),  # MASK_INDEX,
        "instance": (0x01000000, 0)  # MASK_INSTANCE,
    }
    __mask, __childOrSibling, __childOrSiblingSelector = "mask", "childOrSibling", "childOrSiblingSelector"

    def __init__(self, **kwargs):
        super(Selector, self).__setitem__(self.__mask, 0)
        super(Selector, self).__setitem__(self.__childOrSibling, [])
        super(Selector, self).__setitem__(self.__childOrSiblingSelector, [])
        for k in kwargs:
            self[k] = kwargs[k]

    def __setitem__(self, k, v):
        if k in self.__fields:
            super(Selector, self).__setitem__(U(k), U(v))
            super(Selector, self).__setitem__(self.__mask, self[self.__mask] | self.__fields[k][0])
        else:
            raise ReferenceError("%s is not allowed." % k)

    def __delitem__(self, k):
        if k in self.__fields:
            super(Selector, self).__delitem__(k)
            super(Selector, self).__setitem__(self.__mask, self[self.__mask] & ~self.__fields[k][0])

    def clone(self):
        kwargs = dict((k, self[k]) for k in self
                      if k not in [self.__mask, self.__childOrSibling, self.__childOrSiblingSelector])
        selector = Selector(**kwargs)
        for v in self[self.__childOrSibling]:
            selector[self.__childOrSibling].append(v)
        for s in self[self.__childOrSiblingSelector]:
            selector[self.__childOrSiblingSelector].append(s.clone())
        return selector

    def child(self, **kwargs):
        self[self.__childOrSibling].append("child")
        self[self.__childOrSiblingSelector].append(Selector(**kwargs))
        return self

    def sibling(self, **kwargs):
        self[self.__childOrSibling].append("sibling")
        self[self.__childOrSiblingSelector].append(Selector(**kwargs))
        return self

    child_selector, from_parent = child, sibling


def rect(top=0, left=0, bottom=100, right=100):
    return {"top": top, "left": left, "bottom": bottom, "right": right}


def intersect(rect1, rect2):
    top = rect1["top"] if rect1["top"] > rect2["top"] else rect2["top"]
    bottom = rect1["bottom"] if rect1["bottom"] < rect2["bottom"] else rect2["bottom"]
    left = rect1["left"] if rect1["left"] > rect2["left"] else rect2["left"]
    right = rect1["right"] if rect1["right"] < rect2["right"] else rect2["right"]
    return left, top, right, bottom


def point(x=0, y=0):
    return {"x": x, "y": y}


class Adb(object):
    def __init__(self, serial=None, adb_server_host=None, adb_server_port=None):
        self.__adb_cmd = None

        self.default_serial = serial if serial else os.environ.get("ANDROID_SERIAL", None)
        self.adb_server_host = str(adb_server_host if adb_server_host else 'localhost')
        self.adb_server_port = str(adb_server_port if adb_server_port else '5037')
        self.adbHostPortOptions = []
        if self.adb_server_host not in ['localhost', '127.0.0.1']:
            self.adbHostPortOptions += ["-H", self.adb_server_host]
        if self.adb_server_port != '5037':
            self.adbHostPortOptions += ["-P", self.adb_server_port]

            # levyzhang
            # print "self.default_serial: " + str(self.default_serial)
            # print "self.adb_server_host: " + str(self.adb_server_host)
            # print "self.adb_server_port: " + str(self.adb_server_port)
            # print "self.adbHostPortOptions: " + str(self.adbHostPortOptions)

    def adb(self):
        # levyzhang
        return

        if self.__adb_cmd is None:
            if "ANDROID_HOME" in os.environ:
                filename = "adb.exe" if os.name == 'nt' else "adb"
                adb_cmd = os.path.join(os.environ["ANDROID_HOME"], "platform-tools", filename)
                if not os.path.exists(adb_cmd):
                    raise EnvironmentError(
                        "Adb not found in $ANDROID_HOME path: %s." % os.environ["ANDROID_HOME"])
            else:
                # levyzhang
                if False:
                    import distutils
                    if "spawn" not in dir(distutils):
                        import distutils.spawn
                    adb_cmd = distutils.spawn.find_executable("adb")
                    if adb_cmd:
                        adb_cmd = os.path.realpath(adb_cmd)
                    else:
                        raise EnvironmentError("$ANDROID_HOME environment not set.")
                else:
                    filename = "wadb.exe" if os.name == 'nt' else "wadb"
                    adb_cmd = os.path.join(os.getcwd(), "../../wadb", filename)
                    logger.info ("The adb being used is " + str(adb_cmd))
            self.__adb_cmd = adb_cmd
        return self.__adb_cmd

    def cmd(self, *args, **kwargs):
        # levyzhang
        return

        '''adb command, add -s serial by default. return the subprocess.Popen object.'''
        serial = self.device_serial()

        if serial:
            if " " in serial:  # TODO how to include special chars on command line
                serial = "'%s'" % serial

            # levyzhang
            if False:
                return self.raw_cmd(*["-s", serial] + list(args))
            else:
                return self.raw_cmd(*args)
        else:
            return self.raw_cmd(*args)

    def raw_cmd(self, *args):
        # levyzhang
        return

        '''adb command. return the subprocess.Popen object.'''
        cmd_line = [self.adb()] + self.adbHostPortOptions + list(args)
        logger.info (cmd_line)
        if os.name != "nt":
            cmd_line = [" ".join(cmd_line)]
        return subprocess.Popen(cmd_line, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def device_serial(self):
        # levyzhang
        return

        if not self.default_serial:
            devices = self.devices()
            if devices:
                if len(devices) is 1:
                    self.default_serial = list(devices.keys())[0]
                else:
                    raise EnvironmentError("Multiple devices attached but default android serial not set.")
            else:
                raise EnvironmentError("Device not attached.")
        return self.default_serial

    def devices(self):
        # levyzhang
        return

        '''get a dict of attached devices. key is the device serial, value is device name.'''
        out = self.raw_cmd("devices").communicate()[0].decode("utf-8")
        match = "List of devices attached"
        index = out.find(match)
        if index < 0:
            raise EnvironmentError("adb is not working.")
        return dict([s.split("\t") for s in out[index + len(match):].strip().splitlines() if s.strip()])

    def forward(self, local_port, device_port):
        # levyzhang
        return

        '''adb port forward. return 0 if success, else non-zero.'''
        return self.cmd("forward", "tcp:%d" % local_port, "tcp:%d" % device_port).wait()

    def forward_list(self):
        '''adb forward --list'''
        # levyzhang
        return
        '''list all forward socket connections.'''

        version = self.version()
        if int(version[1]) <= 1 and int(version[2]) <= 0 and int(version[3]) < 31:
            raise EnvironmentError("Low adb version.")
        lines = self.raw_cmd("forward", "--list").communicate()[0].decode("utf-8").strip().splitlines()
        return [line.strip().split() for line in lines]

    def version(self):
        # levyzhang
        return

        '''adb version'''
        match = re.search(r"(\d+)\.(\d+)\.(\d+)", self.raw_cmd("version").communicate()[0].decode("utf-8"))
        return [match.group(i) for i in range(4)]


_init_local_port = LOCAL_PORT - 1


def next_local_port(adbHost=None):
    def is_port_listening(port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = s.connect_ex((str(adbHost) if adbHost else '127.0.0.1', port))
        s.close()
        return result == 0

    global _init_local_port
    _init_local_port = _init_local_port + 1 if _init_local_port < 32764 else LOCAL_PORT
    while is_port_listening(_init_local_port):
        _init_local_port += 1
    return _init_local_port


class NotFoundHandler(object):
    '''
    Handler for UI Object Not Found exception.
    It's a replacement of UiAutomator watcher on device side.
    '''

    def __init__(self):
        self.__handlers = collections.defaultdict(lambda: {'on': True, 'handlers': []})

    def __get__(self, instance, type):
        return self.__handlers[instance.adb.device_serial()]


class AutomatorServer(object):
    """start and quit rpc server on device.
    """
    __jar_files = {
        "bundle.jar": "libs/bundle.jar",
        "uiautomator-stub.jar": "libs/uiautomator-stub.jar"
    }

    __apk_files = ["libs/app-uiautomator.apk", "libs/app-uiautomator-test.apk"]

    __sdk = 0

    handlers = NotFoundHandler()  # handler UI Not Found exception

    def __init__(self, serial=None, local_port=None, device_port=None, adb_server_host=None, adb_server_port=None):
        self.uiautomator_process = None
        self.adb = Adb(serial=serial, adb_server_host=adb_server_host, adb_server_port=adb_server_port)
        self.device_port = int(device_port) if device_port else DEVICE_PORT
        if local_port:
            self.local_port = local_port
        else:
            try:  # first we will try to use the local port already adb forwarded
                for s, lp, rp in self.adb.forward_list():
                    if s == self.adb.device_serial() and rp == 'tcp:%d' % self.device_port:
                        self.local_port = int(lp[4:])
                        break
                else:
                    self.local_port = next_local_port(adb_server_host)
            except:
                self.local_port = next_local_port(adb_server_host)

    def push(self):
        base_dir = os.path.dirname(__file__)
        for jar, url in self.__jar_files.items():
            filename = os.path.join(base_dir, url)
            self.adb.cmd("push", filename, "/data/local/tmp/").wait()
        return list(self.__jar_files.keys())

    def install(self):
        base_dir = os.path.dirname(__file__)
        for apk in self.__apk_files:
            self.adb.cmd("install", "-rt", os.path.join(base_dir, apk)).wait()

    @property
    def jsonrpc(self):
        return self.jsonrpc_wrap(timeout=int(os.environ.get("jsonrpc_timeout", 90)))

    @property
    def jsonmonitor(self):
        return self.jsonmonitor_wrap(timeout=int(os.environ.get("jsonrpc_timeout", 90)))

    def jsonrpc_wrap(self, timeout):
        server = self
        ERROR_CODE_BASE = -32000

        def _JsonRPCMethod(url, method, timeout, restart=True):
            _method_obj = JsonRPCMethod(url, method, timeout)

            def wrapper(*args, **kwargs):
                logger.debug("Request start".format(*args))
                URLError = urllib3.exceptions.HTTPError if os.name == "nt" else urllib2.URLError
                try:
                    return _method_obj(*args, **kwargs)
                except (URLError, socket.error, HTTPException) as e:
                    if restart:
                        server.stop()
                        server.start(timeout=30)
                        return _JsonRPCMethod(url, method, timeout, False)(*args, **kwargs)
                    else:
                        raise
                except JsonRPCError as e:
                    if e.code >= ERROR_CODE_BASE - 1:
                        server.stop()
                        server.start()
                        return _method_obj(*args, **kwargs)
                    elif e.code == ERROR_CODE_BASE - 2 and self.handlers['on']:  # Not Found
                        try:
                            self.handlers['on'] = False
                            # any handler returns True will break the left handlers
                            any(handler(self.handlers.get('device', None)) for handler in self.handlers['handlers'])
                        finally:
                            self.handlers['on'] = True
                        return _method_obj(*args, **kwargs)
                    raise
                finally:
                    logger.debug("Request end")

            return wrapper

        return JsonRPCClient(self.rpc_uri,
                             timeout=timeout,
                             method_class=_JsonRPCMethod)

    def jsonmonitor_wrap(self, timeout):
        server = self
        ERROR_CODE_BASE = -32000

        def _JsonRPCMethod(url, method, timeout, restart=True):
            _method_obj = JsonRPCMethod(url, method, timeout)

            def wrapper(*args, **kwargs):
                URLError = urllib3.exceptions.HTTPError if os.name == "nt" else urllib2.URLError
                try:
                    return _method_obj(*args, **kwargs)
                except (URLError, socket.error, HTTPException) as e:
                    if restart:
                        server.stop()
                        server.start(timeout=30)
                        return _JsonRPCMethod(url, method, timeout, False)(*args, **kwargs)
                    else:
                        raise
                except JsonRPCError as e:
                    if e.code >= ERROR_CODE_BASE - 1:
                        server.stop()
                        server.start()
                        return _method_obj(*args, **kwargs)
                    elif e.code == ERROR_CODE_BASE - 2 and self.handlers['on']:  # Not Found
                        try:
                            self.handlers['on'] = False
                            # any handler returns True will break the left handlers
                            any(handler(self.handlers.get('device', None)) for handler in self.handlers['handlers'])
                        finally:
                            self.handlers['on'] = True
                        return _method_obj(*args, **kwargs)
                    raise

            return wrapper

        return JsonRPCClient(self.monitor_uri,
                             timeout=timeout,
                             method_class=_JsonRPCMethod)

    def __jsonrpc(self):
        return JsonRPCClient(self.rpc_uri, timeout=int(os.environ.get("JSONRPC_TIMEOUT", 90)))


    def sdk_version(self):
        '''sdk version of connected device.'''
        if self.__sdk == 0:
            try:
                self.__sdk = int(
                    self.adb.cmd("shell", "getprop", "ro.build.version.sdk").communicate()[0].decode("utf-8").strip())
            except:
                pass
        return self.__sdk

    def start(self, timeout=5):
        # levyzhang
        return
        logger.info ("Start the rpc server.")
        if True:  # self.sdk_version() < 18:
            files = self.push()
            cmd = list(itertools.chain(
                ["shell", "uiautomator", "runtest"],
                files,
                ["-c", "com.github.uiautomatorstub.Stub"]
            ))
        else:
            self.install()
            cmd = ["shell", "am", "instrument", "-w",
                   "com.github.uiautomator.test/android.support.test.runner.AndroidJUnitRunner"]

        self.uiautomator_process = self.adb.cmd(*cmd)
        self.adb.forward(self.local_port, self.device_port)

        while not self.alive and timeout > 0:
            time.sleep(0.1)
            timeout -= 0.1
        if not self.alive:
            raise IOError("RPC server not started!")

    def ping(self):
        try:
            return self.__jsonrpc().ping()
        except:
            return None

    @property
    def alive(self):
        '''Check if the rpc server is alive.'''
        return self.ping() == "pong"

    def stop(self):
        # levyzhang
        return
        logger.info("stop the rpc server.")
        logger.info(self.uiautomator_process)
        '''Stop the rpc server.'''
        if self.uiautomator_process and self.uiautomator_process.poll() is None:
            res = None
            try:
                res = urllib2.urlopen(self.stop_uri)
                self.uiautomator_process.wait()
            except:
                self.uiautomator_process.kill()
            finally:
                if res is not None:
                    res.close()
                self.uiautomator_process = None
        try:
            out = self.adb.cmd("shell", "ps", "-C", "uiautomator").communicate()[0].decode("utf-8").strip().splitlines()
            if out:
                index = out[0].split().index("PID")
                for line in out[1:]:
                    if len(line.split()) > index:
                        self.adb.cmd("shell", "kill", "-9", line.split()[index]).wait()
        except:
            pass

    @property
    def monitor_uri(self):
        return "http://%s:%d/permission" % (self.adb.adb_server_host, self.local_port)

    @property
    def stop_uri(self):
        return "http://%s:%d/stop" % (self.adb.adb_server_host, self.local_port)

    @property
    def rpc_uri(self):
        return "http://%s:%d/jsonrpc/0" % (self.adb.adb_server_host, self.local_port)

    @property
    def screenshot_uri(self):
        return "http://%s:%d/screenshot/0" % (self.adb.adb_server_host, self.local_port)

    def screenshot(self, filename=None, scale=1.0, quality=100):
        if self.sdk_version() >= 18:
            try:
                req = urllib2.Request("%s?scale=%f&quality=%f" % (self.screenshot_uri, scale, quality))
                result = urllib2.urlopen(req, timeout=30)
                if filename:
                    with open(filename, 'wb') as f:
                        f.write(result.read())
                        return filename
                else:
                    return result.read()
            except:
                pass
        return None


# device
class AutomatorDevice(object):
    '''uiautomator wrapper of android device'''

    __orientation = (  # device orientation
                       (0, "natural", "n", 0),
                       (1, "left", "l", 90),
                       (2, "upsidedown", "u", 180),
                       (3, "right", "r", 270)
                       )
    __alias = {
        "width": "displayWidth",
        "height": "displayHeight"
    }

    def __init__(self, serial=None, local_port=None, adb_server_host=None, adb_server_port=None):
        self.server = AutomatorServer(
            serial=serial,
            local_port=local_port,
            adb_server_host=adb_server_host,
            adb_server_port=adb_server_port
        )

    def __call__(self, **kwargs):
        return AutomatorDeviceObject(self, Selector(**kwargs))

    def __getattr__(self, attr):
        '''alias of fields in info property.'''
        info = self.info
        if attr in info:
            return info[attr]
        elif attr in self.__alias:
            return info[self.__alias[attr]]
        else:
            raise AttributeError("%s attribute not found!" % attr)

    @property
    def info(self):
        '''Get the device info.
        {u'model': u'HUAWEI P7-L00', u'displayRotation': 1, u'displaySizeDpY': 360, u'displaySizeDpX': 640, u'brand': u'Huawei', u'displayWidth': 1794,
         u'productName': u'P7-L00', u'release': u'4.4.2',
         u'currentPackageName': u'com.yhkd.jztkc', u'device': u'hwp7', u'sdkInt': 19, u'displayHeight': 1080, u'id': u'HuaweiP7-L00', u'naturalOrientation': False}
        '''
        return self.server.jsonrpc.deviceInfo()

    def click(self, x, y):
        '''click at arbitrary coordinates.'''
        return self.server.jsonrpc.click(x, y)

    def long_click(self, x, y):
        '''long click at arbitrary coordinates.'''
        return self.swipe(x, y, x + 1, y + 1)

    def swipe(self, sx, sy, ex, ey, steps=100):
        return self.server.jsonrpc.swipe(sx, sy, ex, ey, steps)

    def swipePoints(self, points, steps=100):
        ppoints = []
        for p in points:
            ppoints.append(p[0])
            ppoints.append(p[1])
        return self.server.jsonrpc.swipePoints(ppoints, steps)

    def drag(self, sx, sy, ex, ey, steps=100):
        '''Swipe from one point to another point.'''
        return self.server.jsonrpc.drag(sx, sy, ex, ey, steps)

    def dump(self, filename=None, compressed=True, pretty=True):
        '''dump device window and pull to local file.'''
        content = self.server.jsonrpc.dumpWindowHierarchy(compressed, None)
        if content is None:
            logger.info("uiauto dump content is none")
            return
        if filename:
            with open(filename, "wb") as f:
                f.write(content.encode("utf-8"))
        if pretty and "\n " not in content:
            xml_text = xml.dom.minidom.parseString(content.encode("utf-8"))
            content = U(xml_text.toprettyxml(indent='  '))
        return content

    def setdialogtextpattern(self,content):
        if six.PY2:
            content = content.decode("UTF-8")
        logger.info(u"Set text check regex pattern = " + content)
        res = self.server.jsonmonitor.setTextPattern(content)

    def setdialogtextgrouppattern(self,content):
        if six.PY2:
            content = content.decode("UTF-8")
        logger.info(u"Set textgroup check regex pattern = " + content)
        res = self.server.jsonmonitor.setTextMatchGroups(content)

    def setdialogpkgpattern(self, content):
        if six.PY2:
            content = content.decode("UTF-8")
        logger.info(u"Set packge check regex pattern = " + content)
        res = self.server.jsonmonitor.setPackagePattern(content)

    def setPermissionMonitor(self,resume):
        logger.info(u"setPermissionMonitor = " + str(resume))
        res = self.server.jsonmonitor.setPermissionMonitor(resume)

    def screenshot(self, filename, scale=1.0, quality=100):
        '''take screenshot.'''
        result = self.server.screenshot(filename, scale, quality)
        if result:
            return result

        device_file = self.server.jsonrpc.takeScreenshot("screenshot.png",
                                                         scale, quality)
        if not device_file:
            return None
        p = self.server.adb.cmd("pull", device_file, filename)
        p.wait()
        self.server.adb.cmd("shell", "rm", device_file).wait()
        return filename if p.returncode is 0 else None

    def freeze_rotation(self, freeze=True):
        '''freeze or unfreeze the device rotation in current status.'''
        self.server.jsonrpc.freezeRotation(freeze)

    @property
    def orientation(self):
        '''
        orienting the devie to left/right or natural.
        left/l:       rotation=90 , displayRotation=1
        right/r:      rotation=270, displayRotation=3
        natural/n:    rotation=0  , displayRotation=0
        upsidedown/u: rotation=180, displayRotation=2
        '''
        return self.__orientation[self.info["displayRotation"]][1]

    @orientation.setter
    def orientation(self, value):
        '''setter of orientation property.'''
        for values in self.__orientation:
            if value in values:
                # can not set upside-down until api level 18.
                self.server.jsonrpc.setOrientation(values[1])
                break
        else:
            raise ValueError("Invalid orientation.")

    @property
    def last_traversed_text(self):
        '''get last traversed text. used in webview for highlighted text.'''
        return self.server.jsonrpc.getLastTraversedText()

    def clear_traversed_text(self):
        '''clear the last traversed text.'''
        self.server.jsonrpc.clearLastTraversedText()

    @property
    def open(self):
        '''
        Open notification or quick settings.
        Usage:
        d.open.notification()
        d.open.quick_settings()
        '''

        @param_to_property(action=["notification", "quick_settings"])
        def _open(action):
            if action == "notification":
                return self.server.jsonrpc.openNotification()
            else:
                return self.server.jsonrpc.openQuickSettings()

        return _open

    @property
    def handlers(self):
        obj = self

        class Handlers(object):

            def on(self, fn):
                if fn not in obj.server.handlers['handlers']:
                    obj.server.handlers['handlers'].append(fn)
                obj.server.handlers['device'] = obj
                return fn

            def off(self, fn):
                if fn in obj.server.handlers['handlers']:
                    obj.server.handlers['handlers'].remove(fn)

        return Handlers()

    @property
    def watchers(self):
        obj = self

        class Watchers(list):

            def __init__(self):
                for watcher in obj.server.jsonrpc.getWatchers():
                    self.append(watcher)

            @property
            def triggered(self):
                return obj.server.jsonrpc.hasAnyWatcherTriggered()

            def remove(self, name=None):
                if name:
                    obj.server.jsonrpc.removeWatcher(name)
                else:
                    for name in self:
                        obj.server.jsonrpc.removeWatcher(name)

            def reset(self):
                obj.server.jsonrpc.resetWatcherTriggers()
                return self

            def run(self):
                obj.server.jsonrpc.runWatchers()
                return self

        return Watchers()

    def watcher(self, name):
        obj = self

        class Watcher(object):
            def __init__(self):
                self.__selectors = []

            @property
            def triggered(self):
                return obj.server.jsonrpc.hasWatcherTriggered(name)

            def remove(self):
                obj.server.jsonrpc.removeWatcher(name)

            def when(self, **kwargs):
                self.__selectors.append(Selector(**kwargs))
                return self

            def click(self, **kwargs):
                obj.server.jsonrpc.registerClickUiObjectWatcher(name, self.__selectors, Selector(**kwargs))

            @property
            def press(self):
                @param_to_property(
                    "home", "back", "left", "right", "up", "down", "center",
                    "search", "enter", "delete", "del", "recent", "volume_up",
                    "menu", "volume_down", "volume_mute", "camera", "power")
                def _press(*args):
                    obj.server.jsonrpc.registerPressKeyskWatcher(name, self.__selectors, args)

                return _press

        return Watcher()

    @property
    def press(self):
        '''
        press key via name or key code. Supported key name includes:
        home, back, left, right, up, down, center, menu, search, enter,
        delete(or del), recent(recent apps), volume_up, volume_down,
        volume_mute, camera, power.
        Usage:
        d.press.back()  # press back key
        d.press.menu()  # press home key
        d.press(89)     # press keycode
        '''

        @param_to_property(
            key=["home", "back", "left", "right", "up", "down", "center",
                 "menu", "search", "enter", "delete", "del", "recent",
                 "volume_up", "volume_down", "volume_mute", "camera", "power"]
        )
        def _press(key, meta=None):
            if isinstance(key, int):
                return self.server.jsonrpc.pressKeyCode(key, meta) if meta else self.server.jsonrpc.pressKeyCode(key)
            else:
                return self.server.jsonrpc.pressKey(str(key))

        return _press

    def wakeup(self):
        '''turn on screen in case of screen off.'''
        self.server.jsonrpc.wakeUp()

    def sleep(self):
        '''turn off screen in case of screen on.'''
        self.server.jsonrpc.sleep()

    @property
    def screen(self):
        '''
        Turn on/off screen.
        Usage:
        d.screen.on()
        d.screen.off()

        d.screen == 'on'  # Check if the screen is on, same as 'd.screenOn'
        d.screen == 'off'  # Check if the screen is off, same as 'not d.screenOn'
        '''
        devive_self = self

        class _Screen(object):
            def on(self):
                return devive_self.wakeup()

            def off(self):
                return devive_self.sleep()

            def __call__(self, action):
                if action == "on":
                    return self.on()
                elif action == "off":
                    return self.off()
                else:
                    raise AttributeError("Invalid parameter: %s" % action)

            def __eq__(self, value):
                info = devive_self.info
                if "screenOn" not in info:
                    raise EnvironmentError("Not supported on Android 4.3 and belows.")
                if value in ["on", "On", "ON"]:
                    return info["screenOn"]
                elif value in ["off", "Off", "OFF"]:
                    return not info["screenOn"]
                raise ValueError("Invalid parameter. It can only be compared with on/off.")

            def __ne__(self, value):
                return not self.__eq__(value)

        return _Screen()

    @property
    def wait(self):
        '''
        Waits for the current application to idle or window update event occurs.
        Usage:
        d.wait.idle(timeout=1000)
        d.wait.update(timeout=1000, package_name="com.android.settings")
        '''

        @param_to_property(action=["idle", "update"])
        def _wait(action, timeout=1000, package_name=None):
            if timeout / 1000 + 5 > int(os.environ.get("JSONRPC_TIMEOUT", 90)):
                http_timeout = timeout / 1000 + 5
            else:
                http_timeout = int(os.environ.get("JSONRPC_TIMEOUT", 90))
            if action == "idle":
                return self.server.jsonrpc_wrap(timeout=http_timeout).waitForIdle(timeout)
            elif action == "update":
                return self.server.jsonrpc_wrap(timeout=http_timeout).waitForWindowUpdate(package_name, timeout)

        return _wait

    def exists(self, **kwargs):
        '''Check if the specified ui object by kwargs exists.'''
        return self(**kwargs).exists


Device = AutomatorDevice


class AutomatorDeviceUiObject(object):
    '''Represent a UiObject, on which user can perform actions, such as click, set text
    '''

    __alias = {'description': "contentDescription"}

    def __init__(self, device, selector):
        self.device = device
        self.jsonrpc = device.server.jsonrpc
        self.selector = selector

    @property
    def exists(self):
        '''check if the object exists in current window.'''
        return self.jsonrpc.exist(self.selector)

    def __getattr__(self, attr):
        '''alias of fields in info property.'''
        info = self.info
        if attr in info:
            return info[attr]
        elif attr in self.__alias:
            return info[self.__alias[attr]]
        else:
            raise AttributeError("%s attribute not found!" % attr)

    @property
    def info(self):
        '''ui object info.'''
        return self.jsonrpc.objInfo(self.selector)

    def set_text(self, text):
        '''set the text field.'''
        if text in [None, ""]:
            return self.jsonrpc.clearTextField(self.selector)  # TODO no return
        else:
            if isinstance(text, str):
                text = imapUtf7.encode(text.decode("utf-8"))
            else:
                text = imapUtf7.encode(text)
            return self.jsonrpc.setText(self.selector, text)

    def clear_text(self):
        '''clear text. alias for set_text(None).'''
        self.set_text(None)

    def get_text(self):
        return self.jsonrpc.getText(self.selector)

    @property
    def click(self):
        '''
        click on the ui object.
        Usage:
        d(text="Clock").click()  # click on the center of the ui object
        d(text="OK").click.wait(timeout=3000) # click and wait for the new window update
        d(text="John").click.topleft() # click on the topleft of the ui object
        d(text="John").click.bottomright() # click on the bottomright of the ui object
        '''

        @param_to_property(action=["tl", "topleft", "br", "bottomright", "wait"])
        def _click(action=None, timeout=3000):
            if action is None:
                return self.jsonrpc.click(self.selector)
            elif action in ["tl", "topleft", "br", "bottomright"]:
                return self.jsonrpc.click(self.selector, action)
            else:
                return self.jsonrpc.clickAndWaitForNewWindow(self.selector, timeout)

        return _click

    @property
    def long_click(self):
        '''
        Perform a long click action on the object.
        Usage:
        d(text="Image").long_click()  # long click on the center of the ui object
        d(text="Image").long_click.topleft()  # long click on the topleft of the ui object
        d(text="Image").long_click.bottomright()  # long click on the topleft of the ui object
        '''

        @param_to_property(corner=["tl", "topleft", "br", "bottomright"])
        def _long_click(corner=None):
            info = self.info
            if info["longClickable"]:
                if corner:
                    return self.jsonrpc.longClick(self.selector, corner)
                else:
                    return self.jsonrpc.longClick(self.selector)
            else:
                bounds = info.get("visibleBounds") or info.get("bounds")
                if corner in ["tl", "topleft"]:
                    x = (5 * bounds["left"] + bounds["right"]) / 6
                    y = (5 * bounds["top"] + bounds["bottom"]) / 6
                elif corner in ["br", "bottomright"]:
                    x = (bounds["left"] + 5 * bounds["right"]) / 6
                    y = (bounds["top"] + 5 * bounds["bottom"]) / 6
                else:
                    x = (bounds["left"] + bounds["right"]) / 2
                    y = (bounds["top"] + bounds["bottom"]) / 2
                return self.device.long_click(x, y)

        return _long_click

    @property
    def drag(self):
        '''
        Drag the ui object to other point or ui object.
        Usage:
        d(text="Clock").drag.to(x=100, y=100)  # drag to point (x,y)
        d(text="Clock").drag.to(text="Remove") # drag to another object
        '''

        def to(obj, *args, **kwargs):
            if len(args) >= 2 or "x" in kwargs or "y" in kwargs:
                drag_to = lambda x, y, steps=100: self.jsonrpc.dragTo(self.selector, x, y, steps)
            else:
                drag_to = lambda steps=100, **kwargs: self.jsonrpc.dragTo(self.selector, Selector(**kwargs), steps)
            return drag_to(*args, **kwargs)

        return type("Drag", (object,), {"to": to})()

    def gesture(self, start1, start2, *args, **kwargs):
        '''
        perform two point gesture.
        Usage:
        d().gesture(startPoint1, startPoint2).to(endPoint1, endPoint2, steps)
        d().gesture(startPoint1, startPoint2, endPoint1, endPoint2, steps)
        '''

        def to(obj_self, end1, end2, steps=100):
            ctp = lambda pt: point(*pt) if type(pt) == tuple else pt  # convert tuple to point
            s1, s2, e1, e2 = ctp(start1), ctp(start2), ctp(end1), ctp(end2)
            return self.jsonrpc.gesture(self.selector, s1, s2, e1, e2, steps)

        obj = type("Gesture", (object,), {"to": to})()
        return obj if len(args) == 0 else to(None, *args, **kwargs)

    @property
    def pinch(self):
        '''
        Perform two point gesture from edge to center(in) or center to edge(out).
        Usages:
        d().pinch.In(percent=100, steps=10)
        d().pinch.Out(percent=100, steps=100)
        '''

        @param_to_property(in_or_out=["In", "Out"])
        def _pinch(in_or_out="Out", percent=100, steps=50):
            if in_or_out in ["Out", "out"]:
                return self.jsonrpc.pinchOut(self.selector, percent, steps)
            elif in_or_out in ["In", "in"]:
                return self.jsonrpc.pinchIn(self.selector, percent, steps)

        return _pinch

    @property
    def swipe(self):
        '''
        Perform swipe action. if device platform greater than API 18, percent can be used and value between 0 and 1
        Usages:
        d().swipe.right()
        d().swipe.left(steps=10)
        d().swipe.up(steps=10)
        d().swipe.down()
        d().swipe("right", steps=20)
        d().swipe("right", steps=20, percent=0.5)
        '''

        @param_to_property(direction=["up", "down", "right", "left"])
        def _swipe(direction="left", steps=10, percent=1):
            if percent == 1:
                return self.jsonrpc.swipe(self.selector, direction, steps)
            else:
                return self.jsonrpc.swipe(self.selector, direction, percent, steps)

        return _swipe

    @property
    def wait(self):
        '''
        Wait until the ui object gone or exist.
        Usage:
        d(text="Clock").wait.gone()  # wait until it's gone.
        d(text="Settings").wait.exists() # wait until it appears.
        '''

        @param_to_property(action=["exists", "gone"])
        def _wait(action, timeout=3000):
            if timeout / 1000 + 5 > int(os.environ.get("JSONRPC_TIMEOUT", 90)):
                http_timeout = timeout / 1000 + 5
            else:
                http_timeout = int(os.environ.get("JSONRPC_TIMEOUT", 90))
            method = self.device.server.jsonrpc_wrap(
                timeout=http_timeout
            ).waitUntilGone if action == "gone" else self.device.server.jsonrpc_wrap(timeout=http_timeout).waitForExists
            return method(self.selector, timeout)

        return _wait


class AutomatorDeviceNamedUiObject(AutomatorDeviceUiObject):
    def __init__(self, device, name):
        super(AutomatorDeviceNamedUiObject, self).__init__(device, name)

    def child(self, **kwargs):
        return AutomatorDeviceNamedUiObject(
            self.device,
            self.jsonrpc.getChild(self.selector, Selector(**kwargs))
        )

    def sibling(self, **kwargs):
        return AutomatorDeviceNamedUiObject(
            self.device,
            self.jsonrpc.getFromParent(self.selector, Selector(**kwargs))
        )


class AutomatorDeviceObject(AutomatorDeviceUiObject):
    '''Represent a generic UiObject/UiScrollable/UiCollection,
    on which user can perform actions, such as click, set text
    '''

    def __init__(self, device, selector):
        super(AutomatorDeviceObject, self).__init__(device, selector)

    def child(self, **kwargs):
        '''set childSelector.'''
        return AutomatorDeviceObject(
            self.device,
            self.selector.clone().child(**kwargs)
        )

    def sibling(self, **kwargs):
        '''set fromParent selector.'''
        return AutomatorDeviceObject(
            self.device,
            self.selector.clone().sibling(**kwargs)
        )

    child_selector, from_parent = child, sibling

    def child_by_text(self, txt, **kwargs):
        if "allow_scroll_search" in kwargs:
            allow_scroll_search = kwargs.pop("allow_scroll_search")
            name = self.jsonrpc.childByText(
                self.selector,
                Selector(**kwargs),
                txt,
                allow_scroll_search
            )
        else:
            name = self.jsonrpc.childByText(
                self.selector,
                Selector(**kwargs),
                txt
            )
        return AutomatorDeviceNamedUiObject(self.device, name)

    def child_by_description(self, txt, **kwargs):
        if "allow_scroll_search" in kwargs:
            allow_scroll_search = kwargs.pop("allow_scroll_search")
            name = self.jsonrpc.childByDescription(
                self.selector,
                Selector(**kwargs),
                txt,
                allow_scroll_search
            )
        else:
            name = self.jsonrpc.childByDescription(
                self.selector,
                Selector(**kwargs),
                txt
            )
        return AutomatorDeviceNamedUiObject(self.device, name)

    def child_by_instance(self, inst, **kwargs):
        return AutomatorDeviceNamedUiObject(
            self.device,
            self.jsonrpc.childByInstance(self.selector, Selector(**kwargs), inst)
        )

    @property
    def count(self):
        return self.jsonrpc.count(self.selector)

    def __len__(self):
        return self.count

    def __getitem__(self, index):
        count = self.count
        if index >= count:
            raise IndexError()
        elif count == 1:
            return self
        else:
            selector = self.selector.clone()
            selector["instance"] = index
            return AutomatorDeviceObject(self.device, selector)

    def __iter__(self):
        obj, length = self, self.count

        class Iter(object):

            def __init__(self):
                self.index = -1

            def next(self):
                self.index += 1
                if self.index < length:
                    return obj[self.index]
                else:
                    raise StopIteration()

            __next__ = next

        return Iter()

    def right(self, **kwargs):
        def onrightof(rect1, rect2):
            left, top, right, bottom = intersect(rect1, rect2)
            return rect2["left"] - rect1["right"] if top < bottom else -1

        return self.__view_beside(onrightof, **kwargs)

    def left(self, **kwargs):
        def onleftof(rect1, rect2):
            left, top, right, bottom = intersect(rect1, rect2)
            return rect1["left"] - rect2["right"] if top < bottom else -1

        return self.__view_beside(onleftof, **kwargs)

    def up(self, **kwargs):
        def above(rect1, rect2):
            left, top, right, bottom = intersect(rect1, rect2)
            return rect1["top"] - rect2["bottom"] if left < right else -1

        return self.__view_beside(above, **kwargs)

    def down(self, **kwargs):
        def under(rect1, rect2):
            left, top, right, bottom = intersect(rect1, rect2)
            return rect2["top"] - rect1["bottom"] if left < right else -1

        return self.__view_beside(under, **kwargs)

    def __view_beside(self, onsideof, **kwargs):
        bounds = self.info["bounds"]
        min_dist, found = -1, None
        for ui in AutomatorDeviceObject(self.device, Selector(**kwargs)):
            dist = onsideof(bounds, ui.info["bounds"])
            if dist >= 0 and (min_dist < 0 or dist < min_dist):
                min_dist, found = dist, ui
        return found

    @property
    def fling(self):
        '''
        Perform fling action.
        Usage:
        d().fling()  # default vertically, forward
        d().fling.horiz.forward()
        d().fling.vert.backward()
        d().fling.toBeginning(max_swipes=100) # vertically
        d().fling.horiz.toEnd()
        '''

        @param_to_property(
            dimention=["vert", "vertically", "vertical", "horiz", "horizental", "horizentally"],
            action=["forward", "backward", "toBeginning", "toEnd"]
        )
        def _fling(dimention="vert", action="forward", max_swipes=1000):
            vertical = dimention in ["vert", "vertically", "vertical"]
            if action == "forward":
                return self.jsonrpc.flingForward(self.selector, vertical)
            elif action == "backward":
                return self.jsonrpc.flingBackward(self.selector, vertical)
            elif action == "toBeginning":
                return self.jsonrpc.flingToBeginning(self.selector, vertical, max_swipes)
            elif action == "toEnd":
                return self.jsonrpc.flingToEnd(self.selector, vertical, max_swipes)

        return _fling

    @property
    def scroll(self):
        '''
        Perfrom scroll action.
        Usage:
        d().scroll(steps=50) # default vertically and forward
        d().scroll.horiz.forward(steps=100)
        d().scroll.vert.backward(steps=100)
        d().scroll.horiz.toBeginning(steps=100, max_swipes=100)
        d().scroll.vert.toEnd(steps=100)
        d().scroll.horiz.to(text="Clock")
        '''

        def __scroll(vertical, forward, steps=100):
            method = self.jsonrpc.scrollForward if forward else self.jsonrpc.scrollBackward
            return method(self.selector, vertical, steps)

        def __scroll_to_beginning(vertical, steps=100, max_swipes=1000):
            return self.jsonrpc.scrollToBeginning(self.selector, vertical, max_swipes, steps)

        def __scroll_to_end(vertical, steps=100, max_swipes=1000):
            return self.jsonrpc.scrollToEnd(self.selector, vertical, max_swipes, steps)

        def __scroll_to(vertical, **kwargs):
            return self.jsonrpc.scrollTo(self.selector, Selector(**kwargs), vertical)

        @param_to_property(
            dimention=["vert", "vertically", "vertical", "horiz", "horizental", "horizentally"],
            action=["forward", "backward", "toBeginning", "toEnd", "to"])
        def _scroll(dimention="vert", action="forward", **kwargs):
            vertical = dimention in ["vert", "vertically", "vertical"]
            if action in ["forward", "backward"]:
                return __scroll(vertical, action == "forward", **kwargs)
            elif action == "toBeginning":
                return __scroll_to_beginning(vertical, **kwargs)
            elif action == "toEnd":
                return __scroll_to_end(vertical, **kwargs)
            elif action == "to":
                return __scroll_to(vertical, **kwargs)

        return _scroll


# device = AutomatorDevice()



#-*- coding: UTF-8 -*-

import os
import atexit
from ga2.device.android.uiauto import uiautomator
from ga2.common.utils import *
from ga2.cloud import platformHelper
from ga2.device.android.adbProcess import *
from ga2.common.loggerConfig import logger

class UiautoManager(object):

    __device_uiauto_port=9008

    def __init__(self, serial, host, port=None ):
        self.__serial = serial

        if port is None:
            self.__local_port = os.environ.get("UIAUTOMATOR_PORT", "19008")
        else:
            self.__local_port = port
        if is_cloud_mode():
            response = platformHelper.get_platform_client().platform_forward(self.__device_uiauto_port)
            if response is None:
                return
            else:
                self.__local_port = response["localPort"]
        else:
            self.start_uiautomator_server()
        self.client = uiautomator.AutomatorDevice(serial=self.__serial, local_port=int(self.__local_port),adb_server_host=host, adb_server_port=None)


    def start_uiautomator_server(self, dialoghandler=False):
        adb_forward(self.__local_port, self.__device_uiauto_port, self.__serial)
        file_path = os.path.split(os.path.realpath(__file__))[0]
        if dialoghandler:
            jar_name = "uiautomator-stub-dialoghandler.jar"
            config_path = os.path.abspath(
                os.path.join(file_path, "..",  "libs", "wetest_dialog_config.properties"))
            #write_dialog_config(config_path)
            adb_wait("push {0} /data/local/tmp".format(config_path),self.__serial)
        else:
            jar_name="uiautomator-stub.jar"
        uiautomator_stub_path = os.path.abspath( os.path.join(file_path,"..", "libs", jar_name))
        adb_wait("push " + uiautomator_stub_path + " /data/local/tmp",self.__serial)
        uiautomator_process =adb_nowait(
            "shell uiautomator runtest " + jar_name + " -c com.github.uiautomatorstub.Stub", shell=True)
        atexit.register(adb_kill_process_by_name, "uiautomator", self.__serial)
        atexit.register(adb_remove_forward, self.__local_port, self.__serial)
        time.sleep(1)#wait for uiautomator service normal(not a good way)

    def port(self):
        return self.__local_port

    @property
    def ui_device(self):
        return self.client

    # write pkgname filter , unimplemented yet
    # def __write_dialog_config(self, config_path):
    #     if os.environ.get("PKGNAME") is None:
    #         return
    #     content = ""
    #     # with open(config_path,"r") as f:
    #     #     content=f.read()
    #
    #     textpattern = u'(^(完成|关闭|好|好的|确定|确认|安装|下次再说|暂不删除)$|(.*(?<!不|否)(忽略|允(\s)?许|同意)|继续|稍后|暂不|下一步).*)'
    #     textGroupPattern = u'((建议.*清理)|是否卸载|卸载后)&&&(取消|以后再说|下载再说);是否发送错误报告&&&否;为了给您提供丰富的图书资源&&&取消;简化备份恢复流程&&&(以后再说|下次再说)'
    #     pkgfilter = ""
    #     if os.environ.get('PKGNAME'):
    #         pkgfilter = u'(?=(^(?!' + os.environ.get("PKGNAME") + u"$)))"
    #     pkgattern = pkgfilter + u'^(?!(com.tencent.mm|com.tencent.mqq|com.tencent.mobileqq)$).*'
    #     content += "textpattern=" + textpattern + "\n"
    #     content += "textGroupPattern=" + textGroupPattern + "\n"
    #     content += "pkgattern=" + pkgattern + "\n"
    #     content += "needDialogHandler=true" + "\n"
    #
    #     with open(config_path, "w") as f:
    #         f.write(content)

    # def restart_uiautomator_server(self):





#-*- coding: UTF-8 -*-
import os
import cv2
import time
import logging
import datetime
import numpy as np
from .utils import MatchResult,_getPosition,transparent,judge_SimpleColor,highlight,specialBorder

from . import cv2helper
from .tmplatematcher import TemplateMatcher

logger = logging.getLogger('cvmatcher')
logger.setLevel(logging.INFO)
_filter = logging.Filter(name='cvmatcher')


class FeatureMatcher(object):
    # SIFT参数: FILTER_RATIO为SIFT优秀特征点过滤比例值(0-1范围,建议值0.4-0.6)
    filter_ratio = 0.59
    # SIFT参数: SIFT识别时只找出一对相似特征点时的置信度(confidence)
    one_point_confidence = 0.8


    def _init(self, name):
        # self.result = MatchResult()

        FLANN_INDEX_KDTREE = 1  # bug: flann enums are missing
        FLANN_INDEX_LSH = 6
        chunks = name.split('-')
        norm = None
        # Make sure that there is SIFT module in OpenCV.
        if cv2.__version__.startswith("3."):
            # OpenCV3.x, sift is in contrib module, you need to compile it seperately.
            if chunks[0] == 'sift':
                try:
                    detector = cv2helper.eval2('cv2.xfeatures2d.SIFT_create')(edgeThreshold=10)
                    norm = cv2helper.eval2('cv2.NORM_L2')
                except:
                    print("to use SIFT, you should build contrib with opencv3.0")
                    logger.error("There is no SIFT module in your OpenCV environment !")
            elif chunks[0] == 'surf':
                try:
                    detector = cv2helper.eval2('cv2.xfeatures2d.SURF_create')(800)
                    norm = cv2helper.eval2('cv2.NORM_L2')
                except:
                    print("to useSURF, you should build contrib with opencv3.0")
                    logger.error("There is no SURF module in your OpenCV environment !")
            elif chunks[0] == 'orb':
                try:
                    detector = cv2helper.eval2('cv2.ORB_create')(400)
                    norm = cv2helper.eval2('cv2.NORM_HAMMING')
                except:
                    print("to useORB, you should build contrib with opencv3.0")
                    logger.error("There is no ORB module in your OpenCV environment !")
            else:
                raise ValueError('unknown detector:%s' % name)
        
        else:
            # OpenCV2.x, just use it.
            if chunks[0] == 'sift':
                detector = cv2helper.eval2('cv2.SIFT')(edgeThreshold=10)
                norm = cv2helper.eval2('cv2.NORM_L2')
            elif chunks[0] == 'surf':
                detector = cv2helper.eval2('cv2.SURF')(800)
                norm = cv2helper.eval2('cv2.NORM_L2')
            elif chunks[0] == 'orb':
                detector = cv2helper.eval2('cv2.ORB')(400)
                norm = cv2helper.eval2('cv2.NORM_HAMMING')
            else:
                raise ValueError('unknown detector:%s' % name)
        if norm is None:
            return None, None
        
        if 'flann' in chunks:
            if norm == cv2helper.eval2('cv2.NORM_L2'):
                flann_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
            else:
                flann_params = dict(algorithm=FLANN_INDEX_LSH,
                                    table_number=6,  # 12
                                    key_size=12,  # 20
                                    multi_probe_level=1)  # 2
            matcher = cv2helper.eval2('cv2.FlannBasedMatcher')(flann_params,
                                                               {})  # bug : need to pass empty dict (#1329)
        else:
            matcher = cv2helper.eval2('cv2.BFMatcher')(norm)
        return detector, matcher

    def _getImageData(self, filePath, gray):
        if not os.path.exists(filePath):
            logger.error('file not found:%s' % filePath)
            return None

        imgData = cv2helper.imread(filePath)
        if imgData is None:
            logger.error('invalid image format!')
            return None

        channel = imgData.shape[2]
        if not gray or (channel == 1 and gray):
            return imgData

        imgData = cv2helper.cvtColor(imgData)
        if imgData is None:
            logger.error('failed to covert to gray image')
            return None

        return imgData


    def _getKeyPoints(self, detector,imageData, method='sift-flann'):
        keyPoints, desc = detector.detectAndCompute(imageData, None)
        return keyPoints, desc

    def _filterMatchedPoints(self, kp_scene, desc_scene, kp_query, desc_query, matcher, good_ratio, queryImageWidth):
        # if len(kp_scene) < 2 or len(kp_query) < 2:
        ## logger.error("Not enough feature points in input images !")
        # 匹配两个图片中的特征点集,k=2表示每个特征点取出2个最匹配的对应点:
        # if len(kp_scene) >= 2 and len(kp_query) >= 2:
        matches = matcher.knnMatch(desc_query, desc_scene, k=2)

        # print('matches',len(matches))
        good = []
        # good为特征点初选结果,剔除掉前两名匹配太接近的特征点,不是独特优秀的特征点直接筛除(多目标识别情况直接不适用)
        for m, n in matches:
            if m.distance < good_ratio * n.distance:
                good.append(m)


        # good点需要去除重复的部分,（设定源图像不能有重复点）去重时将src图像中的重复点找出即可
        # 去重策略：允许搜索图像对源图像的特征点映射一对多,不允许多对一重复（即不能源图像上一个点对应搜索图像的多个点）
        good_diff, diff_good_point = [], [[]]
        if good:
            for m in good:

                diff_point = [int(kp_scene[m.trainIdx].pt[0]), int(kp_scene[m.trainIdx].pt[1])]

                if diff_point not in diff_good_point:
                    good_diff.append(m)
                    diff_good_point.append(diff_point)

            good = good_diff

        if good:
            # 根据匹配斜率一致原则去除误匹配点
            k = []
            good_temp = good[:]
            for p in good:
                p_query = kp_query[p.queryIdx].pt
                p_scene = kp_scene[p.trainIdx].pt

                k.append((p_scene[1] - p_query[1]) / (queryImageWidth + p_scene[0] - p_query[0]))

            m_k = np.median(k)
            logger.debug('k in KeyPoints filter is :%s' % m_k)
            for i in range(len(good_temp)):
                if abs(k[i] - m_k) > 0.1:
                    good.remove(good_temp[i])
            del good_temp

        return kp_query, kp_scene, good


    def generate_result(self, middle_point, pypts, confi):
        """
        Format the result: 定义图像识别结果格式
        """
        rect = dict(result=middle_point,
                    rectangle=pypts,
                    confidence=confi)
        return rect

    def _handle_onePoints(self, img_scene, img_query, kp_scene, kp_query, good):
        """sift匹配中只有一对匹配的特征点对的情况."""
        # 识别中心即为该匹配点位置
        pts_query = int(kp_query[good[0].queryIdx].pt[0]), int(kp_query[good[0].queryIdx].pt[1])
        pts_scene = int(kp_scene[good[0].trainIdx].pt[0]), int(kp_scene[good[0].trainIdx].pt[1])
        return self._onePoints(pts_query, pts_scene, img_scene, img_query)

    def _handle_twoPoints(self, img_scene, img_query, kp_scene, kp_query, good):
        """处理两对特征点的情况."""

        pts_query1 = int(kp_query[good[0].queryIdx].pt[0]), int(kp_query[good[0].queryIdx].pt[1])
        pts_query2 = int(kp_query[good[1].queryIdx].pt[0]), int(kp_query[good[1].queryIdx].pt[1])
        pts_scene1 = int(kp_scene[good[0].trainIdx].pt[0]), int(kp_scene[good[0].trainIdx].pt[1])
        pts_scene2 = int(kp_scene[good[1].trainIdx].pt[0]), int(kp_scene[good[1].trainIdx].pt[1])

        return self._twoPoints(pts_query1, pts_query2, pts_scene1, pts_scene2, img_scene, img_query)

    def _handle_threePoints(self, img_scene, img_query, kp_scene, kp_query, good):
        """处理三对特征点的情况."""
        # 拿出query和scene的两个点(点1)和(点2点3的中点),
        # 然后根据两个点原则进行后处理(注意ke_query和kp_scene以及queryIdx和trainIdx):
        pts_query1 = int(kp_query[good[0].queryIdx].pt[0]), int(kp_query[good[0].queryIdx].pt[1])
        pts_query2 = int((kp_query[good[1].queryIdx].pt[0] + kp_query[good[2].queryIdx].pt[0]) / 2), int(
            (kp_query[good[1].queryIdx].pt[1] + kp_query[good[2].queryIdx].pt[1]) / 2)
        pts_scene1 = int(kp_scene[good[0].trainIdx].pt[0]), int(kp_scene[good[0].trainIdx].pt[1])
        pts_scene2 = int((kp_scene[good[1].trainIdx].pt[0] + kp_scene[good[2].trainIdx].pt[0]) / 2), int(
            (kp_scene[good[1].trainIdx].pt[1] + kp_scene[good[2].trainIdx].pt[1]) / 2)

        return self._twoPoints(pts_query1, pts_query2, pts_scene1, pts_scene2, img_scene, img_query)

    def _handle_manyPoints(self, img_scene, img_query, kp_scene, kp_query, good):

        num = len(good)
        pts_query_x = []
        pts_query_y = []
        pts_scene_x = []
        pts_scene_y = []
        for p in range(num):
            pts_query_x.append(kp_query[good[p].queryIdx].pt[0])
            pts_query_y.append(kp_query[good[p].queryIdx].pt[1])
            pts_scene_x.append(kp_scene[good[p].trainIdx].pt[0])
            pts_scene_y.append(kp_scene[good[p].trainIdx].pt[1])
        pts_query1 = int(np.min(pts_query_x)), int(np.min(pts_query_y))
        pts_query2 = int(np.max(pts_query_x)), int(np.max(pts_query_y))
        pts_scene1 = int(np.min(pts_scene_x)), int(np.min(pts_scene_y))
        pts_scene2 = int(np.max(pts_scene_x)), int(np.max(pts_scene_y))

        return self._twoPoints(pts_query1, pts_query2, pts_scene1, pts_scene2, img_scene, img_query)



    def _handle_zeroOnePoints(self, sceneImage, queryImage,**kwargs):
        result,w_h_range = self._template_match(sceneImage, queryImage,**kwargs)
        return result,w_h_range


    def _onePoints(self, pts_query, pts_scene, img_scene, img_query):
        """返回一对对匹配特征点情形下的识别结果."""
        middle_point = [int(pts_scene[0]), int(pts_scene[1])]
        pypts = []
        # 从唯一匹配点扩张出目标区域:(注意整数计算要转成浮点数结果!)
        h, w = img_query.shape[:2]
        h_s, w_s = img_scene.shape[:2]

        # 超出左边界取0, 超出右边界取w_s-1, 超出下边界取0, 超出上边界取h_s-1
        x_min, x_max = max(int(pts_scene[0] - pts_query[0]), 0), min(int(pts_scene[0] + w - pts_query[0]), w_s - 1)
        y_min, y_max = max(int(pts_scene[1] - pts_query[1]), 0), min(int(pts_scene[1] + h - pts_query[1]), h_s - 1)

        # 目标矩形的角点按左上、左下、右下、右上的点序：(x_min,y_min)(x_min,y_max)(x_max,y_max)(x_max,y_min)
        pts = np.float32([[x_min, y_min], [x_min, y_max], [x_max, y_max], [x_max, y_min]]).reshape(-1, 1, 2)
        for npt in pts.astype(int).tolist():
            pypts.append(tuple(npt[0]))

        return middle_point, pypts, [x_min, x_max, y_min, y_max, w, h]

    def _twoPoints(self, pts_query1, pts_query2, pts_scene1, pts_scene2, img_scene, img_query):
        """返回两对匹配特征点情形下的识别结果."""
        # 先算出中心点(在im_source中的坐标)：

        middle_point = [int((pts_scene1[0] + pts_scene2[0]) / 2), int((pts_scene1[1] + pts_scene2[1]) / 2)]
        pypts = []
        # 如果特征点同x轴或同y轴(无论src还是sch中),均不能计算出目标矩形区域来,此时返回值同good=1情形
        if pts_query1[0] == pts_query2[0] or pts_query1[1] == pts_query2[1] or pts_scene1[0] == pts_scene2[0] or \
                pts_scene1[1] == pts_scene2[1]:
            logger.debug('Deal twoPoints as one')
            # confidence = self.one_point_confidence
            # one_match = self.generate_result(middle_point, pypts, confidence)
            # return self._handle_onePoints()
            pts_query = [int((pts_query1[0] + pts_query2[0]) / 2), int((pts_query1[1] + pts_query2[1]) / 2)]
            pts_scene = [int((pts_scene1[0] + pts_scene2[0]) / 2), int((pts_scene1[1] + pts_scene2[1]) / 2)]
            return self._onePoints(pts_query, pts_scene, img_scene, img_query)

        # 计算x,y轴的缩放比例：x_scale、y_scale,从middle点扩张出目标区域:(注意整数计算要转成浮点数结果!)
        h, w = img_query.shape[:2]
        h_s, w_s = img_scene.shape[:2]
        x_scale = abs(1.0 * (pts_scene2[0] - pts_scene1[0]) / (pts_query2[0] - pts_query1[0]))
        y_scale = abs(1.0 * (pts_scene2[1] - pts_scene1[1]) / (pts_query2[1] - pts_query1[1]))
        if x_scale-y_scale>0.3:
            y_scale=min(max(y_scale,1),1.3)
            x_scale =y_scale

        if y_scale - x_scale > 0.3:
            x_scale=min(max(x_scale,1),1.3)
            y_scale =x_scale
        # 得到scale后需要对middle_point进行校正,并非特征点中点,而是映射矩阵的中点。
        query_middle_point = int((pts_query1[0] + pts_query2[0]) / 2), int((pts_query1[1] + pts_query2[1]) / 2)
        middle_point[0] = middle_point[0] - int((query_middle_point[0] - w / 2) * x_scale)
        middle_point[1] = middle_point[1] - int((query_middle_point[1] - h / 2) * y_scale)
        middle_point[0] = max(middle_point[0], 0)  # 超出左边界取0  (图像左上角坐标为0,0)
        middle_point[0] = min(middle_point[0], w_s - 1)  # 超出右边界取w_s-1
        middle_point[1] = max(middle_point[1], 0)  # 超出上边界取0
        middle_point[1] = min(middle_point[1], h_s - 1)  # 超出下边界取h_s-1

        # 计算出来rectangle角点的顺序：左上角->左下角->右下角->右上角, 注意：暂不考虑图片转动
        # 超出左边界取0, 超出右边界取w_s-1, 超出下边界取0, 超出上边界取h_s-1
        x_min, x_max = int(max(middle_point[0] - (w * x_scale) / 2, 0)), int(
            min(middle_point[0] + (w * x_scale) / 2, w_s - 1))
        y_min, y_max = int(max(middle_point[1] - (h * y_scale) / 2, 0)), int(
            min(middle_point[1] + (h * y_scale) / 2, h_s - 1))
        # 目标矩形的角点按左上、左下、右下、右上的点序：(x_min,y_min)(x_min,y_max)(x_max,y_max)(x_max,y_min)
        pts = np.float32([[x_min, y_min], [x_min, y_max], [x_max, y_max], [x_max, y_min]]).reshape(-1, 1, 2)
        for npt in pts.astype(int).tolist():
            pypts.append(tuple(npt[0]))

        return middle_point, pypts, [x_min, x_max, y_min, y_max, w, h]


    def _template_match(self, img_scene, img_query,**kwargs):

        h, w = img_query.shape[:2]
        result = TemplateMatcher().searchImage(img_query, img_scene, similarity=kwargs.get("similarity"), gray=kwargs.get("gray"), zoom=0, scaleMatch=True,is_transparent=0)
        result.method='feature_template'
        rect=result.__getattribute__('rect')
        if rect:
            x_min, y_min, x_max, y_max = rect
            w_h_range=[x_min, x_max, y_min, y_max,w,h]
            return result,w_h_range
        else:

            return result,None

    def _detectResult_check(self, w_h_range):
        """校验识别结果区域大小是否符合常理."""
        x_min, x_max, y_min, y_max, w, h = w_h_range
        tar_width, tar_height = x_max - x_min, y_max - y_min
        # 如果src_img中的矩形识别区域的宽和高的像素数＜5,则判定识别失效。认为提取区域待不可能小于5个像素。(截图一般不可能小于5像素)
        if tar_width < 5 or tar_height < 5:
            logger.error("In src_image, Taget area: width or height < 5 pixel.")
            return False
        # 如果矩形识别区域的宽和高,与sch_img的宽高差距超过2倍(屏幕像素差不可能有2倍),认定为识别错误。
        if tar_width < 0.4 * w or tar_width > 2 * w or tar_height < 0.4 * h or tar_height > 2 * h:
            logger.error("Target area is 2 times bigger or 0.4 times smaller than sch_img.")
            return False

        return True
    def _judge_result(self, img_query, img_detect):
        img_query_mean = []
        img_detect_mean = []

        s_h, s_w = img_detect.shape[:2]
        q_h, q_w = img_query.shape[:2]

        for i in range(1):

            scene_l = img_detect[:, 0:int(s_w / 2)]
            scene_r = img_detect[:, int(s_w / 2):s_w]
            query_l = img_query[:, 0:int(q_w / 2)]
            query_r = img_query[:, int(q_w / 2):q_w]
            img_query_mean.append(np.mean(query_l[:, :, i]))
            img_query_mean.append(np.mean(query_r[:, :, i]))
            img_detect_mean.append(np.mean(scene_l[:, :, i]))
            img_detect_mean.append(np.mean(scene_r[:, :, i]))
        judge_rgb = max(abs(np.array(img_query_mean) - np.array(img_detect_mean)))
        img_query_crop=img_query[int(q_h/4):int(q_h/4*3),int(q_w/4):int(q_w/4*3)]
        img_detect_crop=img_detect[int(s_h/2-q_h/4):int(s_h/2+q_h/4),int(s_w/2-q_w/4):int(s_w/2+q_w/4)]
        img_query_hsv = cv2.cvtColor(img_query_crop, cv2.COLOR_BGR2HSV)
        img_detect_hsv = cv2.cvtColor(img_detect_crop, cv2.COLOR_BGR2HSV)
        img_query_v=img_query_hsv[:,:,2]
        img_detect_v =img_detect_hsv[:,:,2]
        judge_v=abs(np.mean(img_query_v) - np.mean(img_detect_v))

        return judge_rgb,judge_v

    def _judge_result_template(self, img_query, img_detect):
        img_query_mean = []
        img_detect_mean = []

        s_h, s_w = img_detect.shape[:2]
        q_h, q_w = img_query.shape[:2]

        for i in range(1):

            img_query_mean.append(np.mean(img_query[:, :, i]))
            img_detect_mean.append(np.mean(img_detect[:, :, i]))
        judge_rgb = max(abs(np.array(img_query_mean) - np.array(img_detect_mean)))
        img_query_crop=img_query[int(q_h/4):int(q_h/4*3),int(q_w/4):int(q_w/4*3)]
        img_query_hsv = cv2.cvtColor(img_query, cv2.COLOR_BGR2HSV)
        img_detect_hsv = cv2.cvtColor(img_detect, cv2.COLOR_BGR2HSV)
        img_query_v=img_query_hsv[:,:,2]
        img_detect_v =img_detect_hsv[:,:,2]
        judge_v=abs(np.mean(img_query_v) - np.mean(img_detect_v))

        return judge_rgb,judge_v

    def _judge_points(self,kp_query,kp_scene, good):
        pts_query_x = []
        for kp in kp_query:
            pts_query_x.append(kp.pt[0])
        pts_query_distribute = int(np.max(pts_query_x)) -int(np.min(pts_query_x))

        num = len(good)
        pts_scene_x = []
        for p in range(num):
            pts_scene_x.append(kp_scene[good[p].trainIdx].pt[0])

        pts_scene_xmin = int(np.min(pts_scene_x))
        pts_scene_xmax = int(np.max(pts_scene_x))
        pts_scene_distribute=pts_scene_xmax-pts_scene_xmin

        # pts_scene_middle = 0
        if pts_scene_distribute < (pts_query_distribute/4):
            logger.debug('points distribute is err,pts_scene_distribute is smaller than 1/4 of pts_query_distribute!' )
            logger.debug('pts_scene_distribute:%s'%pts_scene_distribute)
            logger.debug('pts_query_distribute:%s' % pts_query_distribute)
            return False
        if pts_scene_distribute > 80:
            for s_p in pts_scene_x:
                if s_p > pts_scene_xmin + pts_scene_distribute*1/3 and s_p<pts_scene_xmax-pts_scene_distribute*1/3:
                    # pts_scene_middle += 1
                    return True
            logger.debug('pts_scene_middle is None,not match!')
            return False

        return True


    def cal_ccoeff_confidence(self, img_scene, img_query):
        """求取两张图片的可信度,使用TM_CCOEFF_NORMED方法."""

        img_scene = cv2.cvtColor(img_scene, cv2.COLOR_BGR2GRAY)
        img_query = cv2.cvtColor(img_query, cv2.COLOR_BGR2GRAY)
        res = cv2.matchTemplate(img_scene, img_query, cv2.TM_CCOEFF_NORMED)
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
        confidence = max_val
        return confidence

    def cal_rgb_confidence(self, img_query_rgb, img_scene_rgb,matched_points):
        """同大小彩图计算相似度."""
        # img_scene_HLS=cv2.cvtColor(img_scene_rgb, cv2.COLOR_BGR2HLS)[2]
        # img_query_HLS=cv2.cvtColor(img_query_rgb,cv2.COLOR_BGR2HLS)[2]
        # binary_threshold = np.mean(img_query_L)
        # # print(binary_threshold)
        # img_scene_gray = cv2.cvtColor(img_scene_rgb, cv2.COLOR_BGR2GRAY)
        # img_query_gray = cv2.cvtColor(img_query_rgb, cv2.COLOR_BGR2GRAY)
        #
        # img_scene_L = cv2.adaptiveThreshold(img_scene_L, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, 5, 10)
        # img_query_L = cv2.adaptiveThreshold(img_query_L, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, 5, 10)
        # cv2helper.show(img_query_rgb)
        # cv2helper.show(img_scene_rgb)
        # # res_temp_l = cv2.matchTemplate(img_scene, img_query, cv2.TM_CCOEFF_NORMED)
        # # min_val_l, max_val_l, min_loc_l, max_loc_l = cv2.minMaxLoc(res_temp_l)
        # retval, img_scene_L = cv2.threshold(img_scene_L, binary_threshold, 255, cv2.THRESH_BINARY)
        # retval, img_query_L = cv2.threshold(img_query_L, binary_threshold, 255, cv2.THRESH_BINARY)
        s_h, s_w = img_scene_rgb.shape[:2]
        q_h, q_w = img_query_rgb.shape[:2]

        scene_l = img_scene_rgb[:, 0:int(s_w / 3)]
        scene_m = img_scene_rgb[:, int(s_w / 3):int(s_w / 3 * 2)]
        scene_r = img_scene_rgb[:, int(s_w / 3 * 2):s_w]

        query_l = img_query_rgb[:, 0:int(q_w / 3)]
        query_m = img_query_rgb[:, int(q_w / 3):int(q_w / 3 * 2)]
        query_r = img_query_rgb[:, int(q_w / 3 * 2):q_w]

        # scene_l = img_scene_rgb[:, 0:int(s_w / 4)]
        # scene_m1 = img_scene_rgb[:, int(s_w / 4):int(s_w / 2)]
        # scene_m2 = img_scene_rgb[:, int(s_w /2 ):int(s_w /4*3)]
        # scene_r = img_scene_rgb[:, int(s_w / 4*3):s_w]
        #
        # query_l = img_query_rgb[:, 0:int(q_w / 4)]
        # query_m1 = img_query_rgb[:, int(q_w / 4):int(q_w /  2)]
        # query_m2 = img_query_rgb[:, int(q_w / 2):int(q_w / 4*3)]
        #
        # query_r = img_query_rgb[:, int(q_w / 4 * 3):q_w]

        # scene_l = img_scene_rgb[:, 0:int(s_w / 2)]
        # scene_r = img_scene_rgb[:, int(s_w / 2):s_w]
        #
        # query_l = img_query_rgb[:, 0:int(q_w / 2)]
        # query_r = img_query_rgb[:, int(q_w /2):q_w]

        res_temp_l = cv2.matchTemplate(scene_l, query_l, cv2.TM_CCOEFF_NORMED)
        min_val_l, max_val_l, min_loc_l, max_loc_l = cv2.minMaxLoc(res_temp_l)
        res_temp_m = cv2.matchTemplate(scene_m, query_m, cv2.TM_CCOEFF_NORMED)
        min_val_m, max_val_m, min_loc_m, max_loc_m = cv2.minMaxLoc(res_temp_m)
        # res_temp_m2 = cv2.matchTemplate(scene_m2, query_m2, cv2.TM_CCOEFF_NORMED)
        # min_val_m2, max_val_m2, min_loc_m2, max_loc_m2 = cv2.minMaxLoc(res_temp_m2)
        res_temp_r = cv2.matchTemplate(scene_r, query_r, cv2.TM_CCOEFF_NORMED)
        min_val_r, max_val_r, min_loc_r, max_loc_r = cv2.minMaxLoc(res_temp_r)
        confidence_min = min(max_val_l, max_val_m, max_val_r)
        confidence_max = max(max_val_l, max_val_m, max_val_r)
        # if matched_points>15:
        #     if confidence_max-confidence_min < 0.2:
        #         confidence=confidence_min
        #     else:
        #         confidence=0
        # else:
        #     if confidence_max-confidence_min < 0.3:
        #         confidence=confidence_min
        #     else:
        #         confidence=0
        confidence=confidence_min
        return confidence

    def _matchFeature(self, sceneImage, queryImage, threshold, rgb=True, good_ratio=filter_ratio, **kwargs):
        feature_result=MatchResult()
        feature_result.method='feature'

        sceneImg = sceneImage.copy()
        queryImg = queryImage.copy()
        queryImageHeight, queryImageWidth = queryImage.shape[:2]
        sceneImageHeight, sceneImageWidth = sceneImage.shape[:2]
        # queryImgH = queryImageHeight
        # queryImgW = queryImageWidth
        # sceneImgH = sceneImageHeight
        # sceneImgW = sceneImageWidth
        method = kwargs.get('method', 'sift-flann')
        position_lock=kwargs.get('position_lock')
        animation=kwargs.get('animation')
        is_word = kwargs.get('is_word')
        is_transparent = kwargs.get('is_transparent')
        is_highlight=kwargs.get('is_highlight')
        # if position_lock==1:
        position = kwargs.get('position')
        logger.debug('templateImg position:%s' % str(position))

        adjustThreshold=kwargs.get('adjustThreshold')
        if position and len(position) == 4:
            logger.info('TemplateImg position is locked,crop Img!')
            crop_position=_getPosition(queryImage,sceneImage, position)
            crop_x_min, crop_y_min, crop_x_max, crop_y_max=crop_position
            sceneImage=sceneImage[crop_y_min:crop_y_max, crop_x_min:crop_x_max]
        else:
            logger.info('TemplateImg position is unlocked')
            crop_x_min, crop_y_min, crop_x_max, crop_y_max = 0, 0, 0, 0

        if position and len(position) == 4:
            res_simpleColor = judge_SimpleColor(queryImage)
            if res_simpleColor:
                logger.info('___res_simpleColor____')
                sim, rect = transparent(queryImage, sceneImage)
                if sim >= 0.6:
                    return feature_result, (rect[0] + crop_x_min, rect[1] + crop_y_min, rect[0] + rect[2] + crop_x_min,
                                        rect[1] + rect[3] + crop_y_min)

        if is_highlight==1 and position and len(position) == 4:
            judge_highlight = highlight(sceneImage, sceneImg)
            if judge_highlight is False:
                judge_highlight2 = highlight(sceneImage, queryImg)
                if judge_highlight2 is False:
                    return feature_result, None

        if is_transparent==1 or is_highlight==1:
            sim,rect=transparent(queryImage,sceneImage)
            logger.debug('****transparent or is_highlight:%s'%sim)
            if position and len(position) == 4:
                if sim>=0.6:
                    return feature_result,(rect[0]+crop_x_min, rect[1]+crop_y_min, rect[0] + rect[2]+crop_x_min, rect[1] + rect[3]+crop_y_min)
            else:
                if sim>=0.8:
                    return feature_result,(rect[0] + crop_x_min, rect[1] + crop_y_min, rect[0] + rect[2] + crop_x_min,rect[1] + rect[3] + crop_y_min)
        # cv2helper.show(sceneImage)
        detector, matcher = self._init(method)
        queryFeaturePoints, queryDesc = self._getKeyPoints(detector,queryImage, method='sift-flann')
        sceneFeaturePoints, sceneDesc = self._getKeyPoints(detector,sceneImage, method='sift-flann')
        good_kp_pairs=[]
        if len(sceneFeaturePoints)>1 and len(queryFeaturePoints)>1:

            filtered_query_kp, filtered_scene_kp, good_kp_pairs = self._filterMatchedPoints(sceneFeaturePoints, sceneDesc,
                                                                                            queryFeaturePoints, queryDesc,
                                                                                            matcher, good_ratio,
                                                                                   queryImageWidth)

        else:
            logger.debug('No sceneFeaturePoints!No queryFeaturePoints!')
            good_kp_pairs==None
        logger.info('Feature match! ')
        resizeTag = False
        resizeM = 1
        templateMatchTag = False
        if is_word == 1 and len(good_kp_pairs)>1 and len(queryFeaturePoints) > 20:
            logger.info('Template image is word,feature points is enough ,need judge_points!')
            pts_scene_middle = self._judge_points(filtered_query_kp, filtered_scene_kp, good_kp_pairs)
            # if  pts_scene_middle==0:
            if not pts_scene_middle:
                logger.debug('pts_scene_middle is None!')
                return feature_result,None

        feature_result.miscellaneous['good_kp_pairs']=len(good_kp_pairs)
        feature_result.miscellaneous['queryFeaturePoints'] = len(queryFeaturePoints)
        logger.debug('queryFeaturePoints is:%s' % len(queryFeaturePoints))
        if good_kp_pairs==None or len(good_kp_pairs) == 0 or len(good_kp_pairs) == 1:
            logger.debug('good_kp_pairs is:%s' % len(good_kp_pairs))
            queryFeaturePoints_T=25
            # if animation==1:
            #     queryFeaturePoints_T=15
            if len(queryFeaturePoints) <= queryFeaturePoints_T:

                templateMatchTag=True
                logger.debug('Not enough feature points,template match!')
                template_result,w_h_range = self._handle_zeroOnePoints(sceneImage, queryImage,**kwargs)
                if w_h_range:

                    templateRect = template_result.rect
                    templateRect=[templateRect[0]+crop_x_min,templateRect[1]+crop_y_min,templateRect[2]+crop_x_min,templateRect[3]+crop_y_min]
                    template_result.rect=templateRect
                    logger.debug('0 points:template_result:%s'%template_result)
                    logger.debug('w_h_range:%s'% w_h_range)
                else:
                    template_result.rect=None
                    logger.debug('Template is also not matched!')
                    return template_result,None
            else:
                logger.debug('Enough feature points, not enough good points, feature is not matched!')
                return feature_result,None


        elif len(good_kp_pairs) == 2:

            logger.debug('good_kp_pairs is:%s' % len(good_kp_pairs))
            # 匹配点对为2,根据点对求出目标区域,据此算出可信度：
            origin_result = self._handle_twoPoints(sceneImage, queryImage, filtered_scene_kp, filtered_query_kp,
                                                   good_kp_pairs)
            if isinstance(origin_result, dict):
                if self.one_point_confidence >= threshold:
                    return feature_result,origin_result
                else:
                    return feature_result,None
            else:
                middle_point, pypts, w_h_range = self._handle_twoPoints(sceneImage, queryImage, filtered_scene_kp,
                                                                        filtered_query_kp, good_kp_pairs)
        elif len(good_kp_pairs) == 3:

            logger.debug('good_kp_pairs is:%s' % len(good_kp_pairs))
            # 匹配点对为3,取出点对,求出目标区域,据此算出可信度：
            origin_result = self._handle_threePoints(sceneImage, queryImage, filtered_scene_kp, filtered_query_kp,
                                                     good_kp_pairs)
            if isinstance(origin_result, dict):
                if self.one_point_confidence >= threshold:
                    return feature_result,origin_result
                else:
                    return feature_result,None
            else:
                middle_point, pypts, w_h_range = self._handle_threePoints(sceneImage, queryImage, filtered_scene_kp,
                                                                          filtered_query_kp, good_kp_pairs)
        else:

            logger.debug('good_kp_pairs is:%s' % len(good_kp_pairs))

            # 匹配点对 >= 4个,使用单矩阵映射求出目标区域,据此算出可信度：
            middle_point, pypts, w_h_range = self._handle_manyPoints(sceneImage, queryImage, filtered_scene_kp,
                                                                     filtered_query_kp, good_kp_pairs)
            # print(w_h_range)

        # 第四步：根据识别区域,求出结果可信度,并将结果进行返回:
        # 对识别结果进行合理性校验: 小于5个像素的,或者缩放超过5倍的,一律视为不合法直接raise.
        result_check = self._detectResult_check(w_h_range)


        if result_check:
            x_min, x_max, y_min, y_max, w, h = w_h_range
            logger.debug('simple result_check is correct!')
        else:
            logger.debug('simple result_check is error!')
            return feature_result,None

        detect_img = sceneImage[y_min:y_max, x_min:x_max]
        if is_highlight == 1:
            judge_highlight = highlight(detect_img, sceneImg)
            if judge_highlight is False:
                judge_highlight2 = highlight(sceneImage, queryImg)
                if judge_highlight2 is False:
                    return feature_result, None

        if len(good_kp_pairs)>20:
            judge_rgb, judge_v = self._judge_result(queryImage, detect_img)
            if judge_v>40 and judge_rgb>40:
                return feature_result, None
            x_min = x_min + crop_x_min
            y_min = y_min + crop_y_min
            x_max = x_max + crop_x_min
            y_max = y_max + crop_y_min
            logger.debug('good points is enough ,no need do judge_result !good_kp_pairs :%s'% len(good_kp_pairs))
            if 'showResult' in kwargs and kwargs['showResult']:
                self._showResult(queryImg, sceneImg, (x_min, y_min, x_max, y_max))
                # self._show_feature_match_result(queryImg, sceneImg, sceneFeaturePoints, queryFeaturePoints,
                # good_kp_pairs,
                # resizeTag, resizeM, (x_min, y_min, x_max, y_max))
            return feature_result,(x_min, y_min, x_max, y_max)



        if is_transparent == 1 or animation == 1 or is_highlight==1:

            x_min = x_min + crop_x_min
            y_min = y_min + crop_y_min
            x_max = x_max + crop_x_min
            y_max = y_max + crop_y_min
            if templateMatchTag:
                if 'showResult' in kwargs and kwargs['showResult']:
                    self._showResult(queryImg, sceneImg, (x_min, y_min, x_max, y_max))
                return template_result, (x_min, y_min, x_max, y_max)
            else:
                if 'showResult' in kwargs and kwargs['showResult']:
                    self._showResult(queryImg, sceneImg, (x_min, y_min, x_max, y_max))

                return feature_result, (x_min, y_min, x_max, y_max)

        else:
            if templateMatchTag:

                judge_rgb, judge_v = self._judge_result_template(queryImage, detect_img)
                logger.debug('start judge_result_template!')
                logger.debug('judge_rgb:%s' % judge_rgb)
                logger.debug('judge_v:%s' % judge_v)

                judge_T=15
                if adjustThreshold==1:
                    logger.debug('Use adjustThreshold!')
                    judge_T=20

                if judge_rgb < judge_T and judge_v < judge_T:
                    x_min = x_min + crop_x_min
                    y_min = y_min + crop_y_min
                    x_max = x_max + crop_x_min
                    y_max = y_max + crop_y_min
                    if 'showResult' in kwargs and kwargs['showResult']:
                        self._showResult(queryImg, sceneImg, (x_min, y_min, x_max, y_max))
                        # self._show_feature_match_result(queryImg, sceneImg, sceneFeaturePoints, queryFeaturePoints,
                        #                                             good_kp_pairs,
                        #                                             resizeTag, resizeM, (x_min, y_min, x_max, y_max))
                    logger.debug('return template_result:%s'%template_result)
                    logger.debug('x_min:%s' % x_min)
                    return template_result,(x_min, y_min, x_max, y_max)
                else:
                    template_result.rect=None
                    return template_result,None
            else:
                judge_rgb, judge_v = self._judge_result(queryImage, detect_img)
                logger.debug('start judge_result:')
                logger.debug('judge_rgb:%s'% judge_rgb)
                logger.debug('judge_v:%s' % judge_v)

                if judge_rgb > 100 or judge_v > 100:
                    return feature_result,None
                judge_T_min = 10
                judge_T_max = 35
                if adjustThreshold == 1:
                    logger.debug('Use adjustThreshold!')
                    judge_T_min = 15
                    judge_T_max = 40

                if (judge_rgb<judge_T_max and judge_v<judge_T_max )or judge_rgb<judge_T_min:
                    x_min = x_min + crop_x_min
                    y_min = y_min + crop_y_min
                    x_max = x_max + crop_x_min
                    y_max = y_max + crop_y_min
                    logger.debug('judge_result is correct, image is matched!')
                    if 'showResult' in kwargs and kwargs['showResult']:
                        self._showResult(queryImg, sceneImg, (x_min, y_min, x_max , y_max ))
                        # self._show_feature_match_result(queryImg, sceneImg, sceneFeaturePoints, queryFeaturePoints,
                        #                                             good_kp_pairs,
                        #                                             resizeTag, resizeM, (x_min, y_min, x_max, y_max))
                    logger.debug('return feature_result:%s' % feature_result)
                    logger.debug('x_min:%s' % x_min)
                    return feature_result,(x_min, y_min, x_max , y_max )
                else:
                    logger.debug('judge_result is error, image is not matched!')
                    return feature_result,None



    def _show_feature_match_result(self, queryImg, sceneImg, sceneFeaturePoints, queryFeaturePoints, kp_pairs,
                                   resizeTag, resizeM, rlt):

        if rlt:
            h2, w2 = sceneImg.shape[:2]
            h1, w1 = queryImg.shape[:2]
            if len(sceneImg.shape) > 2:
                vis = cv2helper.eval2('np.zeros')((max(h1, h2), w1 + w2, 3), cv2helper.eval2('np.uint8'))
            else:
                vis = cv2helper.eval2('np.zeros')((max(h1, h2), w1 + w2), cv2helper.eval2('np.uint8'))
            # print('rlt',rlt)
            vis[:h1, :w1] = queryImg
            vis[:h2, w1:w1 + w2] = sceneImg
            x1 = int(rlt[0] / resizeM)
            y1 = int(rlt[1] / resizeM)
            x2 = int(rlt[2] / resizeM)
            y2 = int(rlt[3] / resizeM)
            cv2helper.eval2('cv2.rectangle')(vis, (x1 + w1, y1), (x2 + w1, y2), (0, 0, 255), 2)

        p1 = cv2helper.eval2('np.int32')([queryFeaturePoints[kpp.queryIdx].pt for kpp in kp_pairs])

        if resizeTag:
            w1 = resizeM * w1
        p2 = cv2helper.eval2('np.int32')([sceneFeaturePoints[kpp.trainIdx].pt for kpp in kp_pairs]) + (w1, 0)
        green = (0, 255, 0)

        for (x1, y1), (x2, y2) in zip(p1, p2):
            if resizeTag:
                x1 = int(x1 / resizeM)
                y1 = int(y1 / resizeM)
                x2 = int(x2 / resizeM)
                y2 = int(y2 / resizeM)
            col = green
            cv2helper.eval2('cv2.circle')(vis, (x1, y1), 2, col, -1)
            cv2helper.eval2('cv2.circle')(vis, (x2, y2), 2, col, -1)

        for (x1, y1), (x2, y2) in zip(p1, p2):
            if resizeTag:
                x1 = int(x1 / resizeM)
                y1 = int(y1 / resizeM)
                x2 = int(x2 / resizeM)
                y2 = int(y2 / resizeM)
            cv2helper.eval2('cv2.line')(vis, (x1, y1), (x2, y2), green)

        cv2helper.show(vis)

    def _showResult(self, queryData, sceneData, rlt):
        templ_h, templ_w = queryData.shape[:2]
        # for pt in pts:
        #     cv2helper.eval2('cv2.rectangle')(sceneData, pt, (pt[0]+templ_w, pt[1]+templ_h), (0,0,255), 2)

        h1, w1 = queryData.shape[:2]
        h2, w2 = sceneData.shape[:2]

        if len(sceneData.shape) > 2:
            vis = cv2helper.eval2('np.zeros')((max(h1, h2), w1 + w2, 3))
        else:
            vis = cv2helper.eval2('np.zeros')((max(h1, h2), w1 + w2))

        vis[:h1, :w1] = queryData
        vis[:h2, w1:w1 + w2] = sceneData

        x1 = rlt[0]
        y1 = rlt[1]
        x2 = rlt[2]
        y2 = rlt[3]
        cv2helper.eval2('cv2.rectangle')(vis, (x1 + w1, y1), (x2 + w1, y2), (0, 0, 255), 2)
        # x1,y1,x2,y2 =self._rect(h1,w1,h2,w2,corners,resizeTag,resizeM)

        cv2.imwrite('vis.jpg', vis)
        tmp = cv2.imread('vis.jpg')
        cv2helper.show(tmp)
        os.remove('vis.jpg')

    def searchImage(self, queryImg, sceneImg, similarity=0.65, gray=False,adjustThreshold=0, **kwargs):
        result=MatchResult()
        result.method = 'feature'
        start = time.time()


        scaleRatio = None
        queryImageData = None
        if isinstance(queryImg, str):
            queryImageData = self._getImageData(queryImg, gray)
        elif cv2helper.isNdarray(queryImg):
            queryImageData = queryImg
        else:
            raise TypeError('unknown type:%s' % type(queryImg))

        trainImageData = None
        if isinstance(sceneImg, str):
            trainImageData = self._getImageData(sceneImg, gray)
        elif cv2helper.isNdarray(sceneImg):
            trainImageData = sceneImg
        else:
            raise TypeError('unknown type:%s' % type(sceneImg))

        if trainImageData is None or queryImageData is None:
            logger.error('No trainImageData or queryImageData')
            end = time.time()
            time_cost = end - start
            result.time = time_cost
            result.parameters = kwargs
            return result


        kwargs.update(similarity=similarity)
        kwargs.update(gray=gray)
        kwargs.update(adjustThreshold=adjustThreshold)
        threshold=0.65
        if adjustThreshold==1:
            self.filter_ratio = 0.65
            threshold = 0.6
        match_result,rlt = self._matchFeature(trainImageData, queryImageData, threshold=threshold, rgb=True,
                                 good_ratio=self.filter_ratio, **kwargs)

        ##固定位置匹配不到————如果检测到黑边，去掉固定位置再次检测
        if rlt is None :

            position = kwargs.get('position')
            if position and len(position) == 4:
                have_specialBorder=specialBorder(trainImageData)
                if have_specialBorder:
                    kwargs['position']=None
                    match_result, rlt = self._matchFeature(trainImageData, queryImageData, threshold=threshold,
                                                           rgb=True,
                                                           good_ratio=self.filter_ratio, **kwargs)
        end = time.time()
        time_cost = end - start
        result.time = time_cost
        if match_result.__getattribute__('method')!='feature':
            logger.debug('____________match_result:%s' % match_result)
            return match_result

        result.rect =rlt
        result.parameters = kwargs
        result.miscellaneous['filter_ratio'] = self.filter_ratio
        result.miscellaneous['threshold'] = threshold
        if 'good_kp_pairs' in match_result.miscellaneous:
            result.miscellaneous['good_kp_pairs']=match_result.miscellaneous['good_kp_pairs']
            result.miscellaneous['queryFeaturePoints'] = match_result.miscellaneous['queryFeaturePoints']
            logger.debug('___________result:%s' % result)
        return  result




# -*- coding:UTF-8 -*-
"""封装了cv2模块和numpy模块的一些图像处理的操作，需依赖cv2lib库
"""
import os
import math
import sys

#todo: support linux

try:
    import cv2
except ImportError:
    print('please install cv2 with pip')
    print('pip install opencv-python')
    sys.exit(1)

try:
    import numpy as np
except ImportError:
    print('please install cv2 with pip')
    print('pip install numpy')
    sys.exit(1)

LOAD_IMAGE_COLOR = 1 #cv2.CV_LOAD_IMAGE_COLOR
LOAD_IMAGE_GRAYSCALE = 0 #cv2.CV_LOAD_IMAGE_GRAYSCALE

COLOR_BGR2GRAY = 6 #cv2.COLOR_BGR2GRAY
COLOR_RGB2GRAY = 7 #cv2.COLOR_RGB2GRAY

def isNdarray(data):
    return isinstance(data, np.ndarray)

def imread(filePath, flags=LOAD_IMAGE_COLOR):
    if os.path.exists(filePath):
        return cv2.imread(filePath, flags)
    return None

def rotate(src, rotation):
    """围绕中心点旋转
    
    :type src: cvmat格式
    :param src: 图像数据
    :type rotation: int
    :param rotation: 当前旋转角度
    """
    if rotation <= 0:
        return src

    w = src.shape[1]
    h = src.shape[0]
    
    if w>h and (rotation==90 or rotation==270):
        return src
    
    rangle = np.deg2rad(rotation)  # angle in radians
    # now calculate new image width and height
    nw = (abs(np.sin(rangle) * h) + abs(np.cos(rangle) * w)) 
    nh = (abs(np.cos(rangle) * h) + abs(np.sin(rangle) * w)) 
    # ask OpenCV for the rotation matrix
    rot_mat = cv2.getRotationMatrix2D((nw * 0.5, nh * 0.5), rotation, 1.0)
    # calculate the move from the old center to the new center combined
    # with the rotation
    rot_move = np.dot(rot_mat, np.array([(nw - w) * 0.5, (nh - h) * 0.5, 0]))
    # the move only affects the translation, so update the translation
    # part of the transform
    rot_mat[0, 2] += rot_move[0]
    rot_mat[1, 2] += rot_move[1]
    return cv2.warpAffine(src, 
                          rot_mat, 
                          (int(math.ceil(nw)), int(math.ceil(nh))), 
                          flags=cv2.INTER_LANCZOS4)

def imdecode(src, flags=LOAD_IMAGE_COLOR):
    i = np.frombuffer(src, dtype='uint8')
    return cv2.imdecode(i, flags)

def imencode(postfix, image, params=[]):
    ok, data = cv2.imencode(postfix, image, params)
    if ok:
        return data
    return None

def imwrite(filePath, src):
    return cv2.imwrite(filePath, src)
        
def cvtColor(src, flags=COLOR_BGR2GRAY):
    return cv2.cvtColor(src, flags)
    
def filter2D(src, ddepth, kernel, *args):
    return cv2.filter2D(src, ddepth, kernel, *args)
    
def roi(src, rc):
    l, t, r, b = rc
    h,w = src.shape[:2]
    if (r-l) >= w:
        return src
    if (b-t) >= h:
        return src

    return src[t:b, l:r].copy()
    # data = cv2.cv.GetSubRect(cv2.cv.fromarray(src), (l, t, r - l, b - t))
    # if data:
    #     return np.asarray(data)
    # return None

def circle(src, pt, radius=5, color=(0, 0, 255), thickness=2):
    cv2.circle(src, pt, radius, color, thickness)
    return src

def rectangle(src, pt1, pt2, color=(0, 0, 255), thickness=2):
    cv2.rectangle(src, pt1, pt2, color, thickness)
    return src

def line(src, ptFrom, ptTo, color=(0, 0, 255), thickness=1):
    cv2.line(src, ptFrom, ptTo, color, thickness)
    return src

def resize(src, newWidth, newHeight):
    newImage = cv2.resize(src, (newWidth, newHeight))
    return newImage

def frombuffer(buff, dtype='uint8'):
    return np.frombuffer(buff, dtype=dtype)

def fromstring(string, dtype=np.uint8):
    return np.frombuffer(string, dtype=dtype)
    
def reshape(src, w, h, channels):
    return np.reshape(src, (h,w,channels))
     
def matchTemplate(image, template, method=cv2.TM_CCOEFF_NORMED):
    return cv2.matchTemplate(image, template, method) 

def minMaxLoc(loc):
    return cv2.minMaxLoc(loc)

def eval2(name):
    return eval(name)

def show(src, title='show'):
    cv2.namedWindow(title,2)
    cv2.imshow(title, src)
    cv2.waitKey()
    cv2.destroyAllWindows()
        
def npwhere(condition, *args):
    return np.where(condition, *args)

def createMsgImage(msg, width, height):
    """创建显示信息的图片"""
    img = np.zeros((height, width, 3), np.uint8)
    img[:, :] = (255, 255, 255)
    top = 20
    left = 20
    lines = msg.split('\n')
    row = len(lines)
    for i in range(row):
        line = lines[i]
        cv2.putText(img, line, (left, top + i * 25), cv2.FONT_HERSHEY_COMPLEX, 0.5, (0, 0, 0), 1)

    return imencode('.jpg',img)

if __name__ == '__main__':
    print((eval2('np.ndarray')))
    





import sys
from ga2.common.loggerConfig import *
import subprocess
import threading

class CmdExecuter:

    '''
    execute a command by shell and return its output
    '''
    @staticmethod
    def execute_and_wait(cmdline):
        logger.info("executeAndWait: "+cmdline)
        if 'posix' in sys.builtin_module_names:
            process = subprocess.Popen(cmdline,shell=True,stdout=subprocess.PIPE, stderr=subprocess.STDOUT, preexec_fn=os.setsid)
        else:
            process = subprocess.Popen(cmdline, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        retlines=[]
        while True:
            line = process.stdout.readline()
            if not line :
                break
            logger.debug(line.strip())
            retlines.append(line.strip())

        return retlines

    '''
    execute a command by shell and return the thread instance
    '''
    @staticmethod
    def execute_no_wait(cmdline):
        t = threading.Thread(target=CmdExecuter.execute_and_wait, args=(cmdline,))
        t.setDaemon(True)
        t.start()
        logger.info("execute_no_wait end")
        return t







from ga2.common.cmdExecuter import CmdExecuter


class IMobileDevice():

    def __init__(self, udid):
        self.udid = udid

    def install(self, ipapath):
        cmd = "ideviceinstaller -i " + ipapath
        if self.udid != None:
            cmd = " -u " + self.udid
        CmdExecuter.execute_and_wait(cmd)

    @staticmethod
    def get_device_list():
        cmd = "idevice_id -l "
        return CmdExecuter.execute_and_wait(cmd)

    def uninstall(self, bundleid):
        cmd = "ideviceinstaller -U " + bundleid
        if self.udid != None:
            cmd = " -u " + self.udid
        CmdExecuter.execute_and_wait(cmd)



# -*- coding: UTF-8 -*-

import logging
from ga2.device.iOS import wda

logger = logging.getLogger(__name__)


class WdaManager(object):
    session_map = {}

    def __init__(self, wdaport, wdahost="127.0.0.1"):
        self.__wdaport = wdaport
        self.wda_client = wda.Client('http://%s:%s' % (wdahost, str(self.__wdaport)))
        self._cur_wda_session = None  # self.wda_client.session(None)

    def get_port(self):
        return self.__wdaport

    def new_session(self, bundleid=None):
        self._cur_wda_session = self.wda_client.session(bundleid)
        self.session_map[bundleid] = self._cur_wda_session
        return self._cur_wda_session

    def get_session(self, bundleid=None):
        if bundleid is None:
            if self._cur_wda_session is None:
                self._cur_wda_session = self.new_session(None)
            return self._cur_wda_session
        elif bundleid in self.session_map.keys():
            return self.session_map[bundleid]
        else:
            logger.error("bundleid:" + bundleid + "s not in session map")
            return None

    def get_top_bundleid(self):
        session = self.wda_client.session()
        bundleid = session.bundle_id
        return bundleid

    def close_session(self, bundleid):
        if bundleid is None:
            self._cur_wda_session.close()
            self._cur_wda_session = None
            self.new_session(None)
            for key, value in list(self.session_map.items()):
                if value == self._cur_wda_session:
                    self.session_map.pop(key)
        elif bundleid in self.session_map.keys():
            session = self.session_map.pop(bundleid)
            if self._cur_wda_session == session:
                self._cur_wda_session = None
            session.close()
        else:
            logger.error("bundleid:" + bundleid + "s not in session map")

#
# class WdaManager(object):
#
#     __instance=None
#     sessionMap={}
#     def __init__(self,wdaport):
#         self.__wdaport = os.environ.get("WDA_LOCAL_PORT", wdaport)
#         self.wda_client = wda.Client('http://localhost:%s' % str(self.__wdaport))
#         self._cur_wda_session=self.wda_client.session(None)
#
#     def get_port(self):
#         return self.__wdaport
#
#     def new_session(self,bundleid):
#         self._cur_wda_session = self.wda_client.session(bundleid)
#         self.sessionMap[bundleid]=self._cur_wda_session
#         return self._cur_wda_session
#
#     def get_session(self,bundleid=None):
#         if bundleid is None:
#             if self._cur_wda_session is None:
#                 self._cur_wda_session = self.new_session(None)
#             return self._cur_wda_session
#         elif bundleid in self.sessionMap.keys():
#             return self.sessionMap[bundleid]
#         else:
#             logger.error("bundleid:"+bundleid + "s not in session map")
#             return None
#
#
#
#     def close_session(self,bundleid=None):
#         if bundleid is None:
#             self._cur_wda_session.close()
#             self._cur_wda_session = None
#             self.new_session(None)
#             for key,value in self.sessionMap.items():
#                 if value == self._cur_wda_session:
#                     self.sessionMap.pop(key)
#         elif bundleid in self.sessionMap.keys():
#             session=self.sessionMap.pop(bundleid)
#             session.close()
#         else:
#             logger.error("bundleid:" + bundleid + "s not in session map")
#



from ga2.common.loggerConfig import logger
from ga2.common.utils import callLog
from ga2.device.deviceOrientation import DeviceOrientation
from ga2.common.utils import *
from ga2.cloud.reporter import Reporter
from ga2.common.WebDriverWait import WebDriverWait
from ga2.uiimage.tmplatematcher import TemplateMatcher
# from ga2.uiimage.featurematcher import FeatureMatcher

import cv2


class AutomationHelperImage:

    @staticmethod
    def rescale(rect, width_src, height_src, width_target, height_target):
        ret = list(rect)
        if (height_src - width_src) * (height_target - width_target) < 0:
            ret[0] = rect[0] * 1.0 / width_src
            ret[2] = rect[2] * 1.0 / width_src
            ret[1] = rect[1] * 1.0 / height_src
            ret[3] = rect[3] * 1.0 / height_src
        else:
            ret[0] = rect[0] * 1.0 / width_src
            ret[2] = rect[2] * 1.0 / width_src
            ret[1] = rect[1] * 1.0 / height_src
            ret[3] = rect[3] * 1.0 / height_src
        return ret

    @staticmethod
    def rotate(rect, target_orientation):
        ret = list(rect)
        if target_orientation == DeviceOrientation.PORTRAIT:
            return ret
        elif target_orientation == DeviceOrientation.LANSCAPE:
            ret[0] = rect[1]
            ret[1] = 1.0 - rect[0]
            ret[2] = rect[3]
            ret[3] = 1.0 - rect[2]
        elif target_orientation == DeviceOrientation.PORTRAIT_UPSIDEDOWN:
            ret[0] = 1.0 - rect[0]
            ret[1] = 1.0 - rect[1]
            ret[2] = 1.0 - rect[2]
            ret[3] = 1.0 - rect[3]
        elif target_orientation == DeviceOrientation.LANDSCAPERIGHT:
            ret[0] = 1.0 - rect[1]
            ret[1] = rect[0]
            ret[2] = 1.0 - rect[3]
            ret[3] = rect[2]
        return ret

    @staticmethod
    def trans_rect_to_device(screenimg, rect, device, portrait):
        display_size = device.display_size()
        logger.debug("trans_rect_to_device : display_size {}x{}".format(display_size[0],display_size[1]))
        ret = AutomationHelperImage.rescale(rect, screenimg.shape[1], screenimg.shape[0], display_size[0],
                                            display_size[1])
        logger.debug("scaled result : {}".format(ret))
        if portrait:  #the screenimg has been portraited
            ret = AutomationHelperImage.rotate(ret,  device.orientation())

        return ret

    @staticmethod
    def get_result_by_template(device, param):
        template_path = param["path"]
        portrait = param["portrait"] if "portrait" in param else True
        height = param["height"] if "height" in param else None

        # featureMatcher = FeatureMatcher()
        try:
            screenimg = device.screenshot(portrait=portrait, height=height)
            if screenimg is None:
                return None
            template = cv2.imread(template_path, cv2.IMREAD_COLOR)
            if template is None:
                return None
            if screenimg.shape[2] != template.shape[2]:  # remove alpha value
                screenimg = screenimg[:, :, :3]
            print(screenimg.shape[1], screenimg.shape[0])
            result = TemplateMatcher().searchImage(templateImage=template, sceneImage=screenimg)
            if result and result.rect:
                rect = AutomationHelperImage.trans_rect_to_device(screenimg, result.rect, device, portrait)
                logger.debug("template match result : {} trans to {} ".format(result.rect, rect))
                return rect

            return result.rect if result and result.rect else None
        except Exception as e:
            logger.exception(e)
            return None

        # print(result)

    @staticmethod
    def wait_element_by_image_template(device, param, timeout):
        if device:
            bound = None
            try:
                bound = WebDriverWait(timeout, 2).until(AutomationHelperImage.get_result_by_template, device, param)
                print(bound)  # (sx,sy,dx,dy)
            except Exception as e:
                logger.warn("element wait timeout:" + str(param))
            return bound
        return None

    @staticmethod
    def handle_element_by_image_template(device, param, operator, **kwargs):
        if device:
            rect = AutomationHelperImage.get_result_by_template(device, param)
            if rect is None:
                logger.error("image element is not found : " + str(param))
                return None
            target_pos = ((rect[0] + rect[2]) / 2, (rect[1] + rect[3]) / 2)
            if is_cloud_mode():
                (width, height) = device.display_size()
                Reporter().screenshot_with_mark(width, height, target_pos[0], target_pos[1])
            operator(target_pos[0], target_pos[1], **kwargs)
            return rect
        return None





from setuptools import setup,find_packages
import os
import io

def get_build_env_var_by_name(flag_name):
    
    flag_set = False
    try:
        flag_set = bool(int(os.environ.get('ENABLE_' + flag_name.upper(), flag_set)))
    except Exception:
        pass
    
    if not flag_set:
        try:
            flag_set = bool(int(open(flag_name + ".enabled").read(1)))
        except Exception:
            pass

    return flag_set


def main():
    with io.open("README.md", encoding="utf-8") as fh:
        long_description = fh.read()
    packge_name = 'gautomator2'
    requires = [
                    'requests',
                    'enum34',
                    'urllib3',
                    'pyyaml',
                    'six',
                    'opencv-python',
               ],

    build_contrib = get_build_env_var_by_name("contrib")
    PACKAGES = ["ga2"] + ["%s.%s" % ("ga2", i) for i in find_packages("ga2")]
    if build_contrib:
        print("ga2_contrib will be included...")
        PACKAGES_CONTRIB = ["ga2_contrib"] + ["%s.%s" % ("ga2_contrib", i) for i in find_packages("ga2_contrib")]
        PACKAGES += PACKAGES_CONTRIB
        packge_name = 'gautomator2-contrib'
    print(PACKAGES)
    print(packge_name)
    setup(
        name=packge_name,
        version='0.1.9',
        description='Python Automation Test Framework for Android/iOS  Games/Apps',
        long_description=long_description,
        long_description_content_type="text/markdown",
        keywords=[
            'gautomator2',
            'gautomator 2.0',
            'automation test',
            'mobile automation'
        ],
        author='xavierhan',
        author_email='xavierhan@tencent.com',
        url='http://git.code.oa.com/gautomator2',
    #    packages=[
    #        'ga2',
    #        'ga2.common'
    #    ],
        packages=PACKAGES,
       

        license='Apache 2.0',

        install_requires=requires,
        classifiers = [
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Quality Assurance',
            'Topic :: Software Development :: Testing',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8'
         ],
        python_requires = ">2.6, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    )



if __name__ == '__main__':
    main()



#-*- coding: UTF-8 -*-
__author__ = 'minhuaxu'

class Command(object):
    """
    Defines constants for the standard WebDriver commands.

    While these constants have no meaning in and of themselves, they are
    used to marshal commands through a service that implements WebDriver's
    remote wire protocol:

        https://github.com/SeleniumHQ/selenium/wiki/JsonWireProtocol

    """

    # Keep in sync with org.openqa.selenium.remote.DriverCommand

    TASK_GET = "task"
    TASK_POST = "task"



#-*- coding: UTF-8 -*-
__author__ = 'minhuaxu'

from .exceptions import *

class ErrorCode(object):
    """
    Error codes defined in the WebDriver wire protocol.
    """
    # Keep in sync with org.openqa.selenium.remote.ErrorCodes and errorcodes.h
    SUCCESS = 0
    UNKNOWN_ERROR = [-1, 'server method error']
    METHOD_NOT_ALLOWED = [500, 'server error']



class ErrorHandler(object):
    """
    Handles errors returned by the WebDriver server.
    """
    def check_response(self, response):
        """
        Checks that a JSON response from the WebDriver does not have an error.

        :Args:
         - response - The JSON response from the WebDriver server as a dictionary
           object.

        :Raises: If the response contains an error message.
        """
        status = response.get('status', None)
        if status is None or status == ErrorCode.SUCCESS:
            return


        message=''
        if isinstance(status, int):
            message = response.get('value', None)

        if 399<status<=510:
            error_info="Error code : {0} ,message {1}".format(status,message)
            raise HttpException(error_info)



from ga2.engine.engine import *
from ga2.engine.unityEngine import *
from ga2.engine.UE4Engine import *

class EngineFactory:
    __engineTypeMap = {"Unity3D": UnityEngine, "UE4": UE4Engine}

    # engine_type  is s string contained in the result of get_sdk_version, to indicate the enginetype
    @staticmethod
    def register_new_engine_type(connector_type):
        EngineFactory.__engineTypeMap[connector_type.engine_type()] = connector_type

    @staticmethod
    def recognize_engine_type(version_info):
        for key in EngineFactory.__engineTypeMap:
            if version_info.engine == key:
                return key
        return None

    @staticmethod
    def create_engine_connector(engine_type, address, port):
        if engine_type not in EngineFactory.__engineTypeMap.keys():
            return None
        return EngineFactory.__engineTypeMap[engine_type](address, port)



# -*- coding: UTF-8 -*-
__author__ = 'minhuaxu'

import logging
import sys
import threading

from ga2.common.wetestExceptions import *
from ga2.engine.protocol import Commands

logger = logging.getLogger(__name__)


class RPCReceiveThread(threading.Thread):
    RESET_SOCKET_FUN = '_wetest_socket_change'

    def __init__(self, socket, funcs, event):
        super(RPCReceiveThread, self).__init__()
        self._socket = socket
        self._funcs = funcs
        self._event = event
        self.running = True
        self._recv_result = None
        self.setDaemon(True)

    def close(self):
        self.running = False
        self._socket.close()

    def run(self):
        while self.running:
            try:
                logger.debug("start receive")
                self._recv_result = self._socket.recv_package()
                self._handle(self._recv_result)
            except Exception as e:
                logger.exception(e)
                self._reconnect()

    def _reconnect(self):
        self._socket.close()
        self._socket.connect()
        self._socket.send_package(Commands.PRC_SET_METHOD, RPCReceiveThread.RESET_SOCKET_FUN)

    def _send_callback_result(self, response):
        try:
            self._socket.send_package(Commands.RPC_METHOD, response)
        except Exception as e:
            logger.warning(str(e))

    def _handle(self, command):
        if command["status"] != 0:
            message = "Error code: " + str(self._recv_result['status']) + " msg: " + self._recv_result['data']
            logger.warn(message)
            return

        if command["cmd"] == Commands.PRC_SET_METHOD:
            # TODO specie Command
            logger.debug("receive set rpc method")
            self._event.set()
        elif command["cmd"] == Commands.RPC_METHOD:
            logger.debug("receive call rpc name = {0},value = {1}".format(command['data']['name'], command['data']['value']))
            result = command['data']
            if result['name'] in self._funcs:
                response = {"seq": result['seq'], "status": 0, "name": result['name']}
                try:
                    response['returnValue'] = self._funcs[result['name']](result['value'])
                except Exception as e:
                    response['returnValue'] = e.message
                    response['status'] = 1
                self._send_callback_result(response)
            else:
                logger.warn("Unknow PRC method = {0}".format(result['name']))
        else:
            logger.warn("Unknow command = {0}".format(command["cmd"]))

    def get_result(self, timeout=10):
        self._event.wait(timeout)
        if self._recv_result:
            if self._recv_result["status"] != 0:
                message = "Error code: " + str(self._recv_result['status']) + " msg: " + self._recv_result['data']
                raise WeTestSDKError(message)
            else:
                return self._recv_result['data']
        else:
            raise WeTestSDKError("Receive Message Error")




from enum import Enum


class DeviceOrientation(Enum):
    PORTRAIT = 0
    LANSCAPE = 1
    PORTRAIT_UPSIDEDOWN = 2
    LANDSCAPERIGHT = 3

WDA_ORIENTATION_MAP = {'PORTRAIT': DeviceOrientation.PORTRAIT,
                           'LANDSCAPE': DeviceOrientation.LANSCAPE,
                           'UIA_DEVICE_ORIENTATION_PORTRAIT_UPSIDEDOWN': DeviceOrientation.PORTRAIT_UPSIDEDOWN,
                           'UIA_DEVICE_ORIENTATION_LANDSCAPERIGHT': DeviceOrientation.LANDSCAPERIGHT}


# -*- coding: UTF-8 -*-
"""
Tencent is pleased to support the open source community by making GAutomator available.
Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.

"""
__author__ = 'alexkan kanchuanqi@gmail.com,minhuaxu wukenaihesos@gmail.com'

import json, socket, struct, time
import threading
import sys
import logging
from ga2.common.wetestExceptions import *

logger = logging.getLogger(__name__)


class SocketClient(object):
    def __init__(self, _host='localhost', _port=27018):
        self.host = _host
        self.port = _port
        self.connect()

    def connect(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        self.socket.connect((self.host, self.port))

    def _send_data(self, data):
        try:
            serialized = json.dumps(data)
        except (TypeError, ValueError) as e:
            raise WeTestInvaildArg('You can only send JSON-serializable data')
        length = len(serialized)
        buff = struct.pack("i", length)
        self.socket.send(buff)
        if six.PY3:
            self.socket.sendall(bytes(serialized, encoding='utf-8'))
        else:
            self.socket.sendall(serialized)

    def close(self):
        try:
            self.socket.close()
        except Exception as e:
            logger.exception(e)

    def recv_package(self):
        length_buffer = self.socket.recv(4)
        if length_buffer:
            total = struct.unpack_from("i", length_buffer)[0]
        else:
            raise WeTestSDKError('recv length is None?')
        view = memoryview(bytearray(total))
        next_offset = 0
        while total - next_offset > 0:
            recv_size = self.socket.recv_into(view[next_offset:], total - next_offset)
            next_offset += recv_size
        try:
            if six.PY3:
                deserialized = json.loads(str(view.tobytes(), encoding='utf-8'))
            else:
                deserialized = json.loads(view.tobytes())
            return deserialized
        except (TypeError, ValueError) as e:
            raise WeTestInvaildArg('Data received was not in JSON format')

    def _recv_data(self):
        deserialized = self.recv_package()
        # print(deserialized)
        if deserialized['status'] != 0:
            message = "Error code: " + str(deserialized['status']) + " msg: " + str(deserialized['data'])
            raise WeTestSDKError(message)
        return deserialized['data']

    def send_package(self, cmd, params=None):
        if not params:
            params = ""
        command = {}
        command["cmd"] = cmd
        command["value"] = params
        for retry in range(0, 2):
            try:
                self._send_data(command)
                return
            except WeTestRuntimeError as e:
                raise e
            except socket.timeout:
                self.socket.close()
                self.connect()
                raise WeTestSDKError("Recv Data From SDK timeout")
            except socket.error as e:
                time.sleep(1)
                print("Retry...{0}".format(e.errno))
                self.socket.close()
                self.connect()
                continue
        raise Exception('Socket Error')

    def send_command(self, cmd, params=None, timeout=20):
        # if params != None and not isinstance(params, dict):
        #     raise Exception('Params should be dict')
        if not params:
            params = ""
        command = {}
        command["cmd"] = cmd
        command["value"] = params
        # print(str(command))
        for retry in range(0, 2):
            try:
                self.socket.settimeout(timeout)
                self._send_data(command)
                ret = self._recv_data()
                return ret
            except WeTestRuntimeError as e:
                raise e
            except socket.timeout:
                self.socket.close()
                self.connect()
                raise WeTestSDKError("Recv Data From SDK timeout")
            except socket.error as e:
                time.sleep(1)
                self.socket.close()
                self.connect()
                continue
        raise Exception('Socket Error')


host = 'localhost'
port = 27018


def get_socket_client():
    if get_socket_client.instance:
        return get_socket_client.instance
    else:
        get_socket_client.instance = SocketClient(host, port)
        return get_socket_client.instance


get_socket_client.instance = None
